import numpy as np
import pandas as pd
import polars as pl

from pyfixest.estimation.estimation import feols, fepois
from pyfixest.report.utils import (
    rename_categoricals,
    rename_event_study_coefs,
)
from pyfixest.utils.utils import capture_context, get_data, ssc


def test_multicol_overdetermined_iv():
    data = get_data()
    fit = feols("Y ~ X2 +  f1| f1 | X1 ~ Z1 + Z2", data=data, ssc=ssc(adj=False))

    assert fit._collin_vars == ["f1"]
    assert fit._collin_vars_z == ["f1"]

    np.testing.assert_allclose(
        fit._beta_hat, np.array([-0.993607, -0.174227], dtype=float), rtol=1e-5
    )
    np.testing.assert_allclose(fit._se, np.array([0.104009, 0.018416]), rtol=1e-5)


def test_polars_input():
    data = get_data()
    data_pl = pl.from_pandas(data)
    fit = feols("Y ~ X1", data=data)
    fit.predict(newdata=data_pl)

    data = get_data(model="Fepois")
    data_pl = pl.from_pandas(data)
    fit = fepois("Y ~ X1", data=data_pl)


def test_integer_XY():
    # Create a random number generator
    rng = np.random.default_rng()

    N = 1000
    X = rng.normal(0, 1, N)
    f = rng.choice([0, 1], N)
    Y = 2 * X + rng.normal(0, 1, N) + f * 2
    Y = np.round(Y).astype(np.int64)
    X = np.round(X).astype(np.int64)

    df = pd.DataFrame({"Y": Y, "X": X, "f": f})

    fit1 = feols("Y ~ X | f", data=df, vcov="iid")
    fit2 = feols("Y ~ X + C(f)", data=df)

    np.testing.assert_allclose(fit1.coef().xs("X"), fit2.coef().xs("X"))


def test_coef_update():
    data = get_data()
    data_subsample = data.sample(frac=0.5)
    m = feols("Y ~ X1 + X2", data=data_subsample)
    new_points_id = np.random.choice(
        list(set(data.index) - set(data_subsample.index)), 5
    )
    X_new, y_new = (
        np.c_[
            np.ones(len(new_points_id)), data.loc[new_points_id][["X1", "X2"]].values
        ],
        data.loc[new_points_id]["Y"].values,
    )
    updated_coefs = m.update(X_new, y_new)
    full_coefs = (
        feols(
            "Y ~ X1 + X2",
            data=data.loc[data_subsample.index.append(pd.Index(new_points_id))],
        )
        .coef()
        .values
    )

    np.testing.assert_allclose(updated_coefs, full_coefs)


def test_coef_update_inplace():
    data = get_data()
    data_subsample = data.sample(frac=0.3)
    m = feols("Y ~ X1 + X2", data=data_subsample)
    new_points_id = np.random.choice(
        list(set(data.index) - set(data_subsample.index)), 5
    )
    X_new, y_new = (
        np.c_[
            data.loc[new_points_id][
                ["X1", "X2"]
            ].values  # only pass columns; let `update` add the intercept
        ],
        data.loc[new_points_id]["Y"].values,
    )
    m.update(X_new, y_new, inplace=True)
    full_coefs = (
        feols(
            "Y ~ X1 + X2",
            data=data.loc[data_subsample.index.append(pd.Index(new_points_id))],
        )
        .coef()
        .values
    )
    np.testing.assert_allclose(m.coef().values, full_coefs)


def test_rename_categoricals():
    coefnames = ["C(var)[T.1]", "C(var)[T.2]", "C(var2)[T.1]", "C(var2)[T.2]"]
    renamed = rename_categoricals(coefnames)
    assert renamed == {
        "C(var)[T.1]": "var::1",
        "C(var)[T.2]": "var::2",
        "C(var2)[T.1]": "var2::1",
        "C(var2)[T.2]": "var2::2",
    }

    # with strings:
    coefnames = ["Intercept", "C(f4)[T.B]", "C(f4)[T.C]"]
    renamed = rename_categoricals(coefnames)
    assert renamed == {
        "Intercept": "Intercept",
        "C(f4)[T.B]": "f4::B",
        "C(f4)[T.C]": "f4::C",
    }

    # with reference levels:
    coefnames = [
        "Intercept",
        "C(f4, contr.treatment(base='A'))[T.B]",
        "C(f4, contr.treatment(base='A'))[T.C]",
    ]
    renamed = rename_categoricals(coefnames)
    assert renamed == {
        "Intercept": "Intercept",
        "C(f4, contr.treatment(base='A'))[T.B]": "f4::B",
        "C(f4, contr.treatment(base='A'))[T.C]": "f4::C",
    }

    # without 'T.' in the categorical notation:
    coefnames = [
        "C(f4)[B]",
        "C(f4)[C]",
    ]
    renamed = rename_categoricals(coefnames)
    assert renamed == {
        "C(f4)[B]": "f4::B",
        "C(f4)[C]": "f4::C",
    }

    # without C() and no 'T.' notation
    coefnames = [
        "f4[B]",
        "f4[C]",
    ]
    renamed = rename_categoricals(coefnames)
    assert renamed == {
        "f4[B]": "f4::B",
        "f4[C]": "f4::C",
    }

    # with categoricals:
    coefnames = ["Intercept", "variable1[T.value1]", "variable1[T.value2]"]
    renamed = rename_categoricals(coefnames)
    assert renamed == {
        "Intercept": "Intercept",
        "variable1[T.value1]": "variable1::value1",
        "variable1[T.value2]": "variable1::value2",
    }

    # Test with labels
    coefnames = ["C(variable1)[T.value1]", "variable2[T.value2]"]
    labels = {"variable1": "var1", "variable2": "var2"}
    renamed = rename_categoricals(coefnames, labels=labels)
    assert renamed == {
        "C(variable1)[T.value1]": "var1::value1",
        "variable2[T.value2]": "var2::value2",
    }

    # Test with custom template
    coefnames = ["C(variable1)[T.value1]", "variable2[T.value2]"]
    template = "{variable}--{value}"
    renamed = rename_categoricals(coefnames, template=template)
    assert renamed == {
        "C(variable1)[T.value1]": "variable1--value1",
        "variable2[T.value2]": "variable2--value2",
    }


def test_rename_event_study_coefs():
    coefnames = [
        "C(rel_year, contr.treatment(base=-1.0))[T.-20.0]",
        "C(rel_year, contr.treatment(base=-1.0))[T.-19.0]",
        "Intercept",
    ]

    renamed = rename_event_study_coefs(coefnames)
    assert renamed == {
        "C(rel_year, contr.treatment(base=-1.0))[T.-20.0]": "rel_year::-20.0",
        "C(rel_year, contr.treatment(base=-1.0))[T.-19.0]": "rel_year::-19.0",
        "Intercept": "Intercept",
    }


def _foo():
    "Simulate a callable for testing context capture behavior."
    ...


def test_context_capture():
    # `_foo` is in caller's stack frame, if should be captured
    # call with -1 to account for adding one more frame inside the function
    context = capture_context(-1)
    assert "_foo" in context

    # `_foo` is in caller's stack frame, but we ask for a deeper stack, `_foo` should not be captured
    context = capture_context(1)
    assert "_foo" not in context

    context = capture_context({})
    assert context == {}

    context = capture_context({"_foo": _foo})
    assert context == {"_foo": _foo}



import numpy as np
import pandas as pd
import pytest

# rpy2 imports
import rpy2.robjects as ro
from rpy2.robjects import pandas2ri
from rpy2.robjects.packages import importr

import pyfixest as pf
from pyfixest.estimation.estimation import feols
from pyfixest.utils.utils import ssc

pandas2ri.activate()

# Core R packages
fixest = importr("fixest")
stats = importr("stats")
base = importr("base")
broom = importr("broom")
car = importr("car")


@pytest.mark.against_r_core
@pytest.mark.parametrize(
    "R",
    [
        np.eye(3),
    ],
)
def test_F_test_single_equation_no_clustering(R):
    # Test R * \beta = 0 with single equation.
    # Generate correlated data for dep_var and treat
    np.random.seed(50)
    n = 3000
    treat = np.random.choice([0, 1], size=n)
    X1 = 1 + np.random.randn(n)
    dep_var = 0.0 + 0.0 * treat + 0.0 * X1 + np.random.randn(n)

    data = pd.DataFrame(
        {
            "dep_var": dep_var,
            "treat": treat,
            "X1": X1,
            "year": np.random.choice(range(2000, 2023), size=n),
        }
    )

    fml = "dep_var ~ treat + X1"
    fit = feols(fml, data, vcov=None, ssc=ssc(adj=False))

    # Wald test
    fit.wald_test(R=R, distribution="F")
    f_stat = fit._f_statistic
    p_stat = fit._p_value

    # Compare with R

    r_fit = stats.lm(fml, data=data)
    r_wald = car.linearHypothesis(r_fit, base.matrix(R, 3, 3), test="F")

    r_fstat = r_wald.rx2("F")[1]
    r_pvalue = r_wald.rx2("Pr(>F)")[1]

    np.testing.assert_allclose(f_stat, r_fstat, rtol=1e-03, atol=1e-02)
    np.testing.assert_allclose(p_stat, r_pvalue, rtol=1e-03, atol=1e-02)


@pytest.mark.against_r_core
@pytest.mark.parametrize(
    "R",
    [
        np.array([1, 0]),
        np.array([1, 2]),
        np.array([1, 1]),
        np.array([-1, 1]),
        # np.eye(2) * 2,
    ],
)
def test_F_test_single_equation(R):
    # Test R * \beta = 0 with single equation.
    data = pd.read_csv("pyfixest/did/data/df_het.csv")
    data = data.iloc[1:3000]
    fml = "dep_var ~ treat"
    fit = feols(fml, data, vcov={"CRV1": "year"}, ssc=ssc(adj=False))

    # Wald test
    fit.wald_test(R=R)
    f_stat = fit._f_statistic
    p_value = fit._p_value

    # Compare with R
    r_fit = fixest.feols(
        ro.Formula(fml), data=data, vcov=ro.Formula("~year"), ssc=fixest.ssc(adj=False)
    )

    r_wald = car.linearHypothesis(r_fit, base.matrix(R, 1, 2), test="F")

    r_fstat = r_wald.rx2("F")[1]
    r_pvalue = r_wald.rx2("Pr(>F)")[1]

    np.testing.assert_allclose(f_stat, r_fstat, rtol=1e-03, atol=1e-03)
    np.testing.assert_allclose(p_value, r_pvalue, rtol=1e-03, atol=1e-03)


@pytest.mark.against_r_core
@pytest.mark.parametrize(
    "seedn",
    [
        50,
        20,
        100,
    ],
)
def test_F_test_multiple_equation(seedn):
    # Test R * \beta = 0 with single equation.

    R = np.eye(3)
    # Generate correlated data for dep_var and treat
    np.random.seed(seedn)
    n = 3000
    treat = np.random.choice([0, 1], size=n)
    X1 = 1 + np.random.randn(n)
    dep_var = 0.0 + 1.0 * treat + 0.0 * X1 + np.random.randn(n)

    data = pd.DataFrame(
        {
            "dep_var": dep_var,
            "treat": treat,
            "X1": X1,
            "year": np.random.choice(range(2000, 2023), size=n),
        }
    )

    fml = "dep_var ~ treat + X1"
    fit = feols(fml, data, vcov={"CRV1": "year"}, ssc=ssc(adj=False))

    # Wald test
    fit.wald_test(R=R)
    f_stat = fit._f_statistic
    p_value = fit._p_value

    r_fit = fixest.feols(
        ro.Formula(fml), data=data, vcov=ro.Formula("~year"), ssc=fixest.ssc(adj=False)
    )

    r_wald = car.linearHypothesis(r_fit, base.matrix(R, 3, 3), test="F")

    r_fstat = r_wald.rx2("F")[1]
    r_pvalue = r_wald.rx2("Pr(>F)")[1]

    np.testing.assert_allclose(f_stat, r_fstat, rtol=1e-03, atol=1e-03)
    np.testing.assert_allclose(p_value, r_pvalue, rtol=1e-03, atol=1e-03)


@pytest.mark.against_r_core
@pytest.mark.parametrize(
    "R, fml",
    [
        (np.eye(3), "dep_var ~ treat + X1"),
        (np.array([[-1, 0, 1], [0, 1, 1]]), "dep_var ~ treat + X1"),
        (np.eye(4), "dep_var ~ treat + X1 + X2"),
    ],
)
def test_F_test_multiple_equations_pvalue(R, fml):
    # Test R * \beta = 0 with multiple equations.
    # In this test, we test p-values that are quite larger than 0.0
    # Generate correlated data for dep_var and treat
    Rsize1 = R.shape[0]
    Rsize2 = R.shape[1]
    np.random.seed(0)
    n = 3000
    treat = np.random.choice([0, 1], size=n)
    X1 = 1 + np.random.randn(n)

    if Rsize2 == 3:
        dep_var = 1.0 + 0.5 * treat + 2.0 * X1 + np.random.randn(n)
        data = pd.DataFrame(
            {
                "dep_var": dep_var,
                "treat": treat,
                "X1": X1,
                "year": np.random.choice(range(2000, 2023), size=n),
            }
        )
    else:
        X2 = 2 + np.random.randn(n)
        dep_var = 1.0 + 0.5 * treat + 2.0 * X1 + 3.0 * X2 + np.random.randn(n)
        data = pd.DataFrame(
            {
                "dep_var": dep_var,
                "treat": treat,
                "X1": X1,
                "X2": X2,
                "year": np.random.choice(range(2000, 2023), size=n),
            }
        )

    fit = feols(fml, data, vcov={"CRV1": "year"}, ssc=ssc(adj=False))

    # Wald test
    fit.wald_test(R=R)
    f_stat = fit._f_statistic

    r_fit = fixest.feols(
        ro.Formula(fml), data=data, vcov=ro.Formula("~year"), ssc=fixest.ssc(adj=False)
    )

    r_wald = car.linearHypothesis(r_fit, base.matrix(R, Rsize1, Rsize2), test="F")

    r_fstat = r_wald.rx2("F")[1]

    np.testing.assert_allclose(f_stat, r_fstat, rtol=1e-02, atol=1e-02)


@pytest.mark.against_r_core
@pytest.mark.parametrize(
    "R, q, fml",
    [
        (np.array([[0, 1, 2], [1, 1, 0]]), np.array([-1.169, -0.104]), "Y ~ X1 + X2"),
        (np.array([[0, 1, 2], [1, 1, 0]]), np.array([-1.4, -0.11]), "Y ~ X1 + X2"),
        (
            np.array([[1, 1, 1, 0], [1, 2, 0, 0]]),
            np.array([-0.200, -1.131]),
            "Y ~ X1 + X2 + Z1",
        ),
    ],
)
def test_wald_test_multiple_equations(R, q, fml):
    # Test R * \beta = q with multiple equations.
    # In this test, we test p-values that are quite larger than 0.0
    # Generate correlated data

    R_nrows = R.shape[0]
    R_ncloumns = R.shape[1]

    data = pf.get_data()
    fit_r = stats.lm(fml, data=data)
    R_r = ro.r.matrix(R.reshape(1, R_nrows * R_ncloumns), nrow=R_nrows, byrow=True)

    fit2 = pf.feols(fml, data=data)
    # Define the hypothesis matrix R
    Rpf = R

    # Define the hypothesis values q (both zero)

    # Perform the Wald test
    fit2.wald_test(R=Rpf, q=q, distribution="chi2")

    r_result = car.linearHypothesis(fit_r, R_r, rhs=ro.FloatVector(q), test="Chisq")

    # Extracting p-value from the result
    wald_stat = fit2._wald_statistic
    p_value = fit2._p_value

    r_wald_stat = r_result[4][1]
    r_p_value = r_result[5][1]

    np.testing.assert_allclose(wald_stat, r_wald_stat, rtol=1e-03, atol=1e-05)
    np.testing.assert_allclose(p_value, r_p_value, rtol=1e-03, atol=1e-05)



from importlib import import_module
from typing import Optional, Union

import matplotlib.pyplot as plt
import numba as nb
import numpy as np
import pandas as pd
import seaborn as sns

# Make lets-plot an optional dependency
try:
    from lets_plot import (
        LetsPlot,
        aes,
        geom_density,
        geom_vline,
        ggplot,
        ggtitle,
        theme_bw,
        xlab,
        ylab,
    )

    _HAS_LETS_PLOT = True
except ImportError:
    _HAS_LETS_PLOT = False

from scipy.stats import norm
from tqdm import tqdm

from pyfixest.estimation.demean_ import demean

# Only setup lets-plot if it's available
if _HAS_LETS_PLOT:
    LetsPlot.setup_html()


def _get_ritest_stats_slow(
    data: pd.DataFrame,
    resampvar: str,
    fml: str,
    type: str,
    reps: int,
    model: str,
    rng: np.random.Generator,
    vcov: Union[str, dict[str, str]],
    clustervar_arr: Optional[np.ndarray] = None,
) -> np.ndarray:
    """
    Compute tests statistics using randomization inference (slow).

    Parameters
    ----------
    data : pd.DataFrame
        The input data set.
    resampvar : str
        The name of the treatment variable.
    fml : str
        The formula of the regression model.
    type : str
        The type of the test statistic. Must be one of 'randomization-c'
        or 'randomization-t'. If 'randomization-c', the statistic
        is the regression coefficient.
        If 'randomization-t', the statistic is the t-statistic.
    reps : int
        The number of repetitions.
    model : str
        The model to estimate. Must be one of 'feols' or 'fepois'.
    rng : np.random.Generator
        The random number generator.
    vcov : str or dict[str, str]
        The type of covarianc estimator. See `feols` or `fepois` for details.
    clustervar_arr : np.ndarray, optional
        Array containing the cluster variable. Defaults to None.

    Returns
    -------
    np.ndarray
        The test statistics. For this algorithm, regression coefficients are returned if
        `type` is 'randomization-c', otherwise t-statistics are returned.

    """
    data_resampled = data.copy()
    fml_update = fml.replace(resampvar, f"{resampvar}_resampled")

    fixest_module = import_module("pyfixest.estimation")
    fit_ = getattr(fixest_module, model)

    resampvar_arr = data_resampled[resampvar].to_numpy()

    ri_stats = np.zeros(reps)

    for i in tqdm(range(reps)):
        D_treat = _resample(
            resampvar_arr=resampvar_arr,
            clustervar_arr=clustervar_arr,
            rng=rng,
            iterations=1,
        ).flatten()

        data_resampled[f"{resampvar}_resampled"] = D_treat

        fixest_fit = fit_(fml_update, data=data_resampled, vcov=vcov)
        if type == "randomization-c":
            ri_stats[i] = fixest_fit.coef().xs(f"{resampvar}_resampled")
        else:
            ri_stats[i] = fixest_fit.tstat().xs(f"{resampvar}_resampled")

    return ri_stats


def _get_ritest_stats_fast(
    Y: np.ndarray,
    X: np.ndarray,
    D: np.ndarray,
    coefnames: list,
    resampvar: str,
    reps: int,
    rng: np.random.Generator,
    weights: np.ndarray,
    clustervar_arr: Optional[np.ndarray] = None,
    fval_df: Optional[pd.DataFrame] = None,
) -> np.ndarray:
    """
    Compute tests statistics using randomization inference (fast).

    Parameters
    ----------
    Y : np.ndarray
        The dependent variable. If the model has fixed effects, the dependent
        variable must be demeaned.
    X : np.ndarray
        The design matrix. If the model has fixed effects, the design matrix
        must be demeaned.
    D : np.ndarray
        The treatment variable.
    coefnames : list
        The names of the coefficients in the regression model.
    resampvar : str
        The name of the treatment variable.
    reps : int
        The number of repetitions.
    rng : np.random.Generator
        The random number generator.
    weights : np.ndarray
        The sample weights.
    clustervar_arr : np.ndarray, optional
        Array containing the cluster variable. Defaults to None.
    fval_df : pd.DataFrame, optional
        The fixed effects. Defaults to None.

    Returns
    -------
    np.ndarray
        The test statistics. For this algorithm, regression coefficients are
        returned.
    """
    X_demean = X
    Y_demean = Y.flatten()

    if fval_df is not None:
        fval = np.zeros_like(fval_df)
        for i, col in enumerate(fval_df.columns):
            fval[:, i] = pd.factorize(fval_df[col])[0]

        if fval.dtype != int:
            fval = fval.astype(int)

        fval = fval.reshape(-1, 1) if fval.ndim == 1 else fval
    else:
        fval = None

    idx = coefnames.index(resampvar)
    bool_idx = np.ones(len(coefnames), dtype=bool)
    bool_idx[idx] = False

    X_demean2 = X_demean[:, bool_idx]

    D = D.reshape(-1, 1) if D.ndim == 1 else D
    X_demean2 = X_demean2.reshape(-1, 1) if X_demean2.ndim == 1 else X_demean2

    resampvar_arr = D

    return _run_ri(
        reps=reps,
        resampvar_arr=resampvar_arr,
        clustervar_arr=clustervar_arr,
        Y_demean=Y_demean,
        X_demean2=X_demean2,
        fval=fval,
        weights=weights,
        rng=rng,
    )


@nb.njit()
def _run_ri(
    reps: int,
    resampvar_arr: np.ndarray,
    rng: np.random.Generator,
    fval: np.ndarray,
    weights: np.ndarray,
    Y_demean: np.ndarray,
    X_demean2: np.ndarray,
    clustervar_arr: Optional[np.ndarray] = None,
) -> np.ndarray:
    """
    Run the randomization inference.

    Parameters
    ----------
    reps : int
        The number of repetitions.
    resampvar_arr : np.ndarray
        Array containing the treatment variable.
    rng : np.random.Generator
        The random number generator.
    fval : np.ndarray
        The fixed effects decoded as integers.
    weights : np.ndarray
        The sample weights.
    X_demean2 : np.ndarray
        The demeaned design matrix.
    Y_demean : np.ndarray
    clustervar_arr : np.ndarray, optional
        Array containing the cluster variable. Defaults to None.

    Returns
    -------
    np.ndarray
        The coefficients of the randomization inference.
        For this algorithm, regression coefficients are
        returned.
    """
    ri_coefs = np.zeros(reps)

    X_demean2 = np.ascontiguousarray(X_demean2)

    for i in range(reps):
        D2 = _resample(
            resampvar_arr=resampvar_arr,
            clustervar_arr=clustervar_arr,
            rng=rng,
            iterations=1,
        )

        D2_demean = demean(D2, fval, weights)[0] if fval is not None else D2

        ri_coefs[i] = lstsq_numba(
            np.concatenate((D2_demean, X_demean2), axis=1), Y_demean
        )[0]

    return ri_coefs


@nb.njit
def _resample(
    resampvar_arr: np.ndarray,
    rng: np.random.Generator,
    iterations: int = 1,
    clustervar_arr: Optional[np.ndarray] = None,
) -> np.ndarray:
    """
    Random resampling of the treatment variable.

    Parameters
    ----------
    resampvar_arr : np.ndarray
        Array containing the treatment variable.
    rng : np.random.Generator
        The random number generator.
    iterations : int, optional
        The number of iterations. Defaults to 1. If bigger than 1,
        return `iterations` resampled treatment variables. Basically,
        this argument allows to vectorize the resampling process.
    clustervar_arr : np.ndarray, optional
        Array containing the cluster variable. Defaults to None.

    Returns
    -------
    np.ndarray
        The resampled treatment variable(s). If `iterations` is bigger
        than 1, the array has shape (N, iterations), where N is the
        number of observations. Otherwise, the array has shape (N,1).
    """
    N = resampvar_arr.shape[0]
    resampvar_values = np.unique(resampvar_arr)

    if clustervar_arr is not None:
        clustervar_values = np.unique(clustervar_arr)
        D_treat = np.zeros((N, iterations))

        for i, _ in enumerate(clustervar_values):
            idx = (clustervar_arr == clustervar_values[i]).flatten()
            D_treat[idx, :] = random_choice(resampvar_values, 1 * iterations, rng)

    else:
        D_treat = random_choice(resampvar_values, N * iterations, rng).reshape(
            (N, iterations)
        )

    return D_treat


@nb.njit
def random_choice(arr: np.ndarray, size: int, rng: np.random.Generator) -> np.ndarray:
    """
    Randomly sample from an array.

    Parameters
    ----------
    arr : np.ndarray
        The array from which to sample.
    size : int
        The number of samples.
    rng : np.random.Generator
        The random number generator.

    Returns
    -------
    np.ndarray
        The sampled array (with replacement) of size `size`.
    """
    n = len(arr)
    result = np.empty(size, dtype=arr.dtype)
    for i in range(size):
        idx = rng.integers(0, n)
        result[i] = arr[idx]
    return result


@nb.njit()
def lstsq_numba(A, B):
    """Implement np.linalg.lstsq(A, B) using SVD decomposition."""
    U, s, VT = np.linalg.svd(A, full_matrices=False)
    c = np.dot(U.T, B)
    w = np.linalg.solve(np.diag(s), c)
    x = np.dot(VT.T, w)
    return x


def _get_ritest_pvalue(
    sample_stat: np.ndarray,
    ri_stats: np.ndarray,
    method: str,
    level: float,
    h0_value: float,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Compute the p-value of the test statistic and
    standard error and CI of the p-value.
    """
    reps = len(ri_stats)
    ci_sides = [0, 1]

    if method == "two-sided":
        probs = np.abs(ri_stats) >= np.abs(sample_stat - h0_value)
    elif method == "greater":
        probs = ri_stats <= sample_stat - h0_value
    elif method == "lower":
        probs = ri_stats >= sample_stat - h0_value
    else:
        raise ValueError(
            "The `method` argument must be one of 'two-sided', 'right', 'left'."
        )

    p_value = probs.mean()
    se_pval = norm.ppf(level) * np.std(probs) / np.sqrt(reps)
    ci_margin = norm.ppf(level) * se_pval
    ci_pval = p_value + np.array([-ci_margin, ci_margin])[ci_sides]

    return p_value, se_pval, ci_pval


def _plot_ritest_pvalue(
    sample_stat: np.ndarray, ri_stats: np.ndarray, plot_backend: str
):
    """Plot the permutation distribution of the test statistic."""
    df = pd.DataFrame({"ri_stats": ri_stats})

    title = "Permutation distribution of the test statistic"
    x_lab = "Test statistic"
    y_lab = "Density"

    if plot_backend == "lets_plot":
        if not _HAS_LETS_PLOT:
            print("lets-plot is not installed. Falling back to matplotlib.")
            plot_backend = "matplotlib"
        else:
            plot = (
                ggplot(df, aes(x="ri_stats"))
                + geom_density(fill="blue", alpha=0.5)
                + theme_bw()
                + geom_vline(xintercept=sample_stat, color="red")
                + ggtitle(title)
                + xlab(x_lab)
                + ylab(y_lab)
            )

        return plot.show()

    elif plot_backend == "matplotlib":
        plt.figure(figsize=(10, 6))
        sns.kdeplot(data=df, x="ri_stats", fill=True, color="blue", alpha=0.5)
        plt.axvline(x=sample_stat, color="red", linestyle="--")
        plt.title(title)
        plt.xlabel(x_lab)
        plt.ylabel(y_lab)
        plt.show()

    else:
        raise ValueError(f"Unsupported plot backend: {plot_backend}")


def _decode_resampvar(resampvar: str) -> tuple[str, float, str, str]:
    """
    Decode the resampling variable.

    Parameters
    ----------
    resampvar : str
        The resampling variable. It can be of the form "var=value",
        "var>value", or "var<value".

    Returns
    -------
    str
        The name of the resampling variable, h0_value (float),
        transformed hypothesis,
        and test_type (two-sided, greater, lower).
    """
    if "=" in resampvar:
        resampvar_, h0_value_str = resampvar.split("=")
        test_type = "two-sided"
        hypothesis = f"{resampvar_}={h0_value_str}"
    elif ">" in resampvar:
        resampvar_, h0_value_str = resampvar.split(">")
        test_type = "greater"
        hypothesis = f"{resampvar_}>{h0_value_str}"
    elif "<" in resampvar:
        resampvar_, h0_value_str = resampvar.split("<")
        test_type = "lower"
        hypothesis = f"{resampvar_}<{h0_value_str}"
    else:
        resampvar_ = resampvar
        test_type = "two-sided"
        hypothesis = f"{resampvar_}=0"
        h0_value_str = "0"

    h0_value_float = float(h0_value_str)

    return resampvar_, h0_value_float, hypothesis, test_type



import numpy as np
import pandas as pd
import pytest
import rpy2.robjects as ro
from rpy2.robjects import pandas2ri

# rpy2 imports
from rpy2.robjects.packages import importr

from pyfixest.estimation.estimation import feols

pandas2ri.activate()

fixest = importr("fixest")
stats = importr("stats")
broom = importr("broom")


@pytest.mark.against_r_core
def test_i():
    df_het = pd.read_csv("pyfixest/did/data/df_het.csv")
    df_het["X"] = np.random.normal(size=len(df_het))

    if (
        "C(rel_year)[T.1.0]"
        in feols("dep_var~i(rel_year, ref = 1.0)", df_het)._coefnames
    ):
        raise AssertionError("C(rel_year)[T.1.0] should not be in the column names.")
    if (
        "C(rel_year)[T.-2.0]"
        in feols("dep_var~i(rel_year,ref=-2.0)", df_het)._coefnames
    ):
        raise AssertionError("C(rel_year)[T.-2.0] should not be in the column names.")

    if (
        "C(rel_year)[T.1.0]:treat"
        in feols("dep_var~i(rel_year, treat, ref=1.0)", df_het)._coefnames
    ):
        raise AssertionError(
            "C(rel_year)[T.1.0]:treat should not be in the column names."
        )
    if (
        "C(rel_year)[T.-2.0]:treat"
        in feols("dep_var~i(rel_year, treat,ref=-2.0)", df_het)._coefnames
    ):
        raise AssertionError(
            "C(rel_year)[T.-2.0]:treat should not be in the column names."
        )

    with pytest.raises(ValueError):
        feols("dep_var~i(rel_year, ref = [1.0, 'a'])", df_het)


@pytest.mark.against_r_core
def test_i_vs_fixest():
    df_het = pd.read_csv("pyfixest/did/data/df_het.csv")
    df_het = df_het[df_het["year"] >= 2010]
    # ------------------------------------------------------------------------ #
    # no fixed effects

    # no references
    fit_py = feols("dep_var~i(treat)", df_het)
    fit_r = fixest.feols(ro.Formula("dep_var~i(treat)"), df_het)
    np.testing.assert_allclose(
        fit_py.coef().values, np.array(fit_r.rx2("coefficients"))
    )

    fit_py = feols("dep_var~i(rel_year)", df_het)
    fit_r = fixest.feols(ro.Formula("dep_var~i(rel_year)"), df_het)
    np.testing.assert_allclose(
        fit_py.coef().values, np.array(fit_r.rx2("coefficients"))
    )

    # with references
    fit_py = feols("dep_var~i(treat, ref = False)", df_het)
    fit_r = fixest.feols(ro.Formula("dep_var~i(treat, ref = FALSE)"), df_het)
    np.testing.assert_allclose(
        fit_py.coef().values, np.array(fit_r.rx2("coefficients"))
    )

    fit_py = feols("dep_var~i(rel_year, ref = 1.0)", df_het)
    fit_r = fixest.feols(ro.Formula("dep_var~i(rel_year, ref = c(1))"), df_het)
    np.testing.assert_allclose(
        fit_py.coef().values, np.array(fit_r.rx2("coefficients"))
    )

    # ------------------------------------------------------------------------ #
    # with fixed effects

    # no references
    fit_py = feols("dep_var~i(treat) | year", df_het)
    fit_r = fixest.feols(ro.Formula("dep_var~i(treat)|year"), df_het)
    np.testing.assert_allclose(
        fit_py.coef().values, np.array(fit_r.rx2("coefficients"))
    )

    fit_py = feols("dep_var~i(rel_year) | year", df_het)
    fit_r = fixest.feols(ro.Formula("dep_var~i(rel_year)|year"), df_het)
    np.testing.assert_allclose(
        fit_py.coef().values, np.array(fit_r.rx2("coefficients"))
    )

    # with references
    fit_py = feols("dep_var~i(treat,ref=False) | year", df_het)
    fit_r = fixest.feols(ro.Formula("dep_var~i(treat, ref = FALSE)|year"), df_het)
    np.testing.assert_allclose(
        fit_py.coef().values, np.array(fit_r.rx2("coefficients"))
    )

    fit_py = feols("dep_var~i(rel_year,ref=1.0) | year", df_het)
    fit_r = fixest.feols(ro.Formula("dep_var~i(rel_year, ref = c(1))|year"), df_het)
    np.testing.assert_allclose(
        fit_py.coef().values, np.array(fit_r.rx2("coefficients"))
    )


@pytest.mark.against_r_core
@pytest.mark.parametrize(
    "fml",
    [
        "dep_var ~ i(state)",
        "dep_var ~ i(state, ref = 1)",
        "dep_var ~ i(state, year)",
        "dep_var ~ i(state, year, ref = 1)",
        "dep_var ~ i(state, year) | state",
        "dep_var ~ i(state, year, ref = 1) | state",
    ],
)
def test_i_interacted_fixest(fml):
    df_het = pd.read_csv("pyfixest/did/data/df_het.csv")
    df_het["X"] = np.random.normal(df_het.shape[0])

    fit_py = feols(fml, df_het)
    fit_r = fixest.feols(ro.Formula(fml), df_het)
    np.testing.assert_allclose(
        fit_py.coef().values, np.array(fit_r.rx2("coefficients"))
    )



import numpy as np
import pytest

import pyfixest as pf


# @pytest.mark.skip(reason="Bug for fweights and heteroskedastic errors.")
def test_fweights_ols():
    "Test that the fweights are correctly implemented for OLS models."
    # Fepois model for discrete Y
    data = pf.get_data(model="Fepois")
    data2_w = (
        data[["Y", "X1"]]
        .groupby(["Y", "X1"])
        .size()
        .reset_index()
        .rename(columns={0: "count"})
    )
    data3_w = (
        data[["Y", "X1", "f1"]]
        .groupby(["Y", "X1", "f1"])
        .size()
        .reset_index()
        .rename(columns={0: "count"})
    )

    fit1 = pf.feols("Y ~ X1", data=data, ssc=pf.ssc(adj=False, cluster_adj=False))
    fit2 = pf.feols(
        "Y ~ X1",
        data=data2_w,
        weights="count",
        weights_type="fweights",
        ssc=pf.ssc(adj=False, cluster_adj=False),
    )

    assert fit1._N == fit2._N, "Number of observations is not the same."

    if False:
        np.testing.assert_allclose(fit1.tidy().values, fit2.tidy().values)

        np.testing.assert_allclose(fit1.vcov("HC1")._vcov, fit2.vcov("HC1")._vcov)
        np.testing.assert_allclose(fit1.vcov("HC2")._vcov, fit2.vcov("HC2")._vcov)
        np.testing.assert_allclose(fit1.vcov("HC3")._vcov, fit2.vcov("HC3")._vcov)

        fit3 = pf.feols("Y ~ X1 | f1", data=data)
        fit4 = pf.feols(
            "Y ~ X1 | f1", data=data3_w, weights="count", weights_type="fweights"
        )
        np.testing.assert_allclose(fit3.tidy().values, fit4.tidy().values)
        np.testing.assert_allclose(
            fit3.vcov({"CRV3": "f1"})._vcov, fit4.vcov({"CRV3": "f1"})._vcov
        )
        np.testing.assert_allclose(fit1.vcov("HC1")._vcov, fit2.vcov("HC1")._vcov)
        np.testing.assert_allclose(fit1.vcov("HC2")._vcov, fit2.vcov("HC2")._vcov)
        np.testing.assert_allclose(fit1.vcov("HC3")._vcov, fit2.vcov("HC3")._vcov)


@pytest.mark.skip(reason="Not implemented yet.")
def test_fweights_iv():
    data = pf.get_data()
    data2_w = (
        data[["Y", "X1", "Z1"]]
        .groupby(["Y", "X1", "Z1"])
        .size()
        .reset_index()
        .rename(columns={0: "count"})
    )

    fit1 = pf.feols("Y ~ 1 | X1 ~ Z1", data=data)
    fit2 = pf.feols(
        "Y ~ 1 | X1 ~ Z1", data=data2_w, weights="count", weights_type="fweights"
    )
    np.testing.assert_allclose(fit1.tidy().values, fit2.tidy().values)

    data3_w = (
        data[["Y", "X1", "Z1", "f1"]]
        .groupby(["Y", "X1", "Z1", "f1"])
        .size()
        .reset_index()
        .rename(columns={0: "count"})
    )

    fit3 = pf.feols("Y ~ 1 | f1 | X1 ~ Z1 ", data=data.dropna(), vcov={"CRV1": "f1"})
    fit4 = pf.feols(
        "Y ~ 1 | f1 | X1 ~ Z1",
        data=data3_w.dropna(),
        weights="count",
        weights_type="fweights",
        vcov={"CRV1": "f1"},
    )
    np.testing.assert_allclose(fit3.tidy().values, fit4.tidy().values)


def test_aweights():
    data = pf.get_data()
    data["weights"] = np.ones(data.shape[0])

    fit1 = pf.feols("Y ~ X1", data=data)
    fit2 = pf.feols("Y ~ X1", data=data, weights_type="aweights")
    fit3 = pf.feols("Y ~ X1", data=data, weights="weights", weights_type="aweights")

    np.testing.assert_allclose(fit1.tidy().values, fit2.tidy().values)
    np.testing.assert_allclose(fit1.tidy().values, fit3.tidy().values)



import numpy as np
import pytest

import pyfixest as pf


def check_absolute_diff(x1, x2, tol, msg=None):
    "Check for absolute differences."
    if isinstance(x1, (int, float)):
        x1 = np.array([x1])
    if isinstance(x2, (int, float)):
        x2 = np.array([x2])
        msg = "" if msg is None else msg

    # handle nan values
    nan_mask_x1 = np.isnan(x1)
    nan_mask_x2 = np.isnan(x2)

    if not np.array_equal(nan_mask_x1, nan_mask_x2):
        raise AssertionError(f"{msg}: NaN positions do not match")

    valid_mask = ~nan_mask_x1  # Mask for non-NaN elements (same for x1 and x2)
    assert np.all(np.abs(x1[valid_mask] - x2[valid_mask]) < tol), msg


fml_list = [
    ("Y ~ X1 + C(f1)", "Y~X1 | f1"),
    ("Y ~ X1 + C(f1) + C(f2)", "Y~X1 | f1 + f2"),
]

fml_ols_vs_gaussian = ["Y ~ X1", "Y ~ X1 + C(f1)", "Y ~ X1 * X2"]


@pytest.mark.parametrize("fml", fml_ols_vs_gaussian)
@pytest.mark.parametrize("inference", ["iid", "hetero", {"CRV1": "f1"}])
@pytest.mark.parametrize("dropna", [True])
def test_ols_vs_gaussian_glm(fml, inference, dropna):
    data = pf.get_data()
    if dropna:
        data = data.dropna()

    fit_ols = pf.feols(fml=fml, data=data, vcov=inference)
    fit_gaussian = pf.feglm(fml=fml, data=data, family="gaussian", vcov=inference)

    check_absolute_diff(
        fit_ols.coef().xs("X1"), fit_gaussian.coef().xs("X1"), tol=1e-10
    )
    check_absolute_diff(fit_ols._weights[0:5], fit_gaussian._weights[0:5], tol=1e-10)
    check_absolute_diff(fit_ols._u_hat[0:5], fit_gaussian._u_hat[0:5], tol=1e-10)
    check_absolute_diff(fit_ols._scores[0, :], fit_gaussian._scores[0, :], tol=1e-10)

    if inference != "iid":
        # iid inference different: follows iid-glm; just the bread and not bread x sigma2
        check_absolute_diff(
            fit_ols.se().xs("X1"), fit_gaussian.se().xs("X1"), tol=1e-10
        )
        check_absolute_diff(fit_ols._vcov[0, 0], fit_gaussian._vcov[0, 0], tol=1e-10)


@pytest.mark.skip("Fixed effects are not yet supported.")
@pytest.mark.parametrize("fml", fml_list)
@pytest.mark.parametrize("family", ["gaussian"])
def test_feols_feglm_internally(fml, family):
    data = pf.get_data()
    data["Y"] = np.where(data["Y"] > 0, 1, 0)

    fml1, fml2 = fml

    fit1 = pf.feglm(
        fml=fml1, data=data, family=family, ssc=pf.ssc(adj=False, cluster_adj=False)
    )
    fit2 = pf.feglm(
        fml=fml2, data=data, family=family, ssc=pf.ssc(adj=False, cluster_adj=False)
    )

    assert fit1.coef().xs("X1") == fit2.coef().xs("X1"), (
        f"Test failed for fml = {fml} and family = gaussian"
    )
    assert fit1.se().xs("X1") == fit2.se().xs("X1"), (
        f"Test failed for fml = {fml} and family = gaussian"
    )
    assert fit1._u_hat[0:5]



import numpy as np
import pandas as pd
import pytest
from formulaic.errors import FactorEvaluationError

import pyfixest as pf
from pyfixest.errors import (
    DuplicateKeyError,
    EndogVarsAsCovarsError,
    InstrumentsAsCovarsError,
    NanInClusterVarError,
    UnderDeterminedIVError,
    VcovTypeNotSupportedError,
)
from pyfixest.estimation.estimation import feols, fepois
from pyfixest.estimation.FormulaParser import FixestFormulaParser
from pyfixest.estimation.multcomp import rwolf
from pyfixest.report.summarize import etable, summary
from pyfixest.utils.dgps import gelbach_data
from pyfixest.utils.utils import get_data, ssc


@pytest.fixture
def data():
    return pf.get_data()


def test_formula_parser2():
    with pytest.raises(DuplicateKeyError):
        FixestFormulaParser("y ~ sw(a, b) +  sw(c, d)| sw(X3, X4))")


def test_formula_parser3():
    with pytest.raises(DuplicateKeyError):
        FixestFormulaParser("y ~ sw(a, b) +  csw(c, d)| sw(X3, X4))")


# def test_formula_parser2():
#    with pytest.raises(FixedEffectInteractionError):
#        FixestFormulaParser('y ~ X1 + X2 | X3:X4')

# def test_formula_parser3():
#    with pytest.raises(CovariateInteractionError):
#        FixestFormulaParser('y ~ X1 + X2^X3')


def test_cluster_na():
    """Test if a nan value in a cluster variable raises an error."""
    data = get_data()
    data = data.dropna()
    data["f3"] = data["f3"].astype("int64")
    data["f3"][5] = np.nan

    with pytest.raises(NanInClusterVarError):
        feols(fml="Y ~ X1", data=data, vcov={"CRV1": "f3"})


def test_cluster_but_no_data():
    """Test if AttributeError if self._data is not stored."""
    data = get_data()
    fit = feols("Y ~ X1", data=data, store_data=False)
    with pytest.raises(AttributeError):
        fit.vcov({"CRV1": "f2"})


def test_error_hc23_fe():
    """
    Test if HC2 & HC3 inference with fixed effects regressions raises an error.

    Notes
    -----
    Currently not supported.
    """
    data = get_data().dropna()

    with pytest.raises(VcovTypeNotSupportedError):
        feols(fml="Y ~ X1 | f2", data=data, vcov="HC2")

    with pytest.raises(VcovTypeNotSupportedError):
        feols(fml="Y ~ X1 | f2", data=data, vcov="HC3")


def test_depvar_numeric():
    """Test if feols() throws an error when the dependent variable is not numeric."""
    data = get_data()
    data["Y"] = data["Y"].astype("str")
    data["Y"] = pd.Categorical(data["Y"])

    with pytest.raises(TypeError):
        feols(fml="Y ~ X1", data=data)


def test_iv_errors():
    data = get_data()

    # under determined
    with pytest.raises(UnderDeterminedIVError):
        feols(fml="Y ~ X1 | Z1 + Z2 ~ 24 ", data=data)
    # instrument specified as covariate
    with pytest.raises(InstrumentsAsCovarsError):
        feols(fml="Y ~ X1 | Z1  ~ X1 + X2", data=data)
    # endogenous variable specified as covariate
    with pytest.raises(EndogVarsAsCovarsError):
        feols(fml="Y ~ Z1 | Z1  ~ X1", data=data)

    # instrument specified as covariate
    # with pytest.raises(InstrumentsAsCovarsError):
    #    fixest.feols('Y ~ X1 | Z1 + Z2 ~ X3 + X4')
    # underdetermined IV
    # with pytest.raises(UnderDeterminedIVError):
    #    fixest.feols('Y ~ X1 + X2 | X1 + X2 ~ X4 ')
    # with pytest.raises(UnderDeterminedIVError):
    #    fixest.feols('Y ~ X1 | Z1 + Z2 ~ X2 + X3 ')
    # CRV3 inference
    with pytest.raises(VcovTypeNotSupportedError):
        feols(fml="Y ~ 1 | Z1 ~ X1 ", vcov={"CRV3": "group_id"}, data=data)
    # wild bootstrap
    with pytest.raises(NotImplementedError):
        feols(fml="Y ~ 1 | Z1 ~ X1 ", data=data).wildboottest(param="Z1", reps=999)
    # multi estimation error
    with pytest.raises(NotImplementedError):
        feols(fml="Y + Y2 ~ 1 | Z1 ~ X1 ", data=data)
    with pytest.raises(NotImplementedError):
        feols(fml="Y  ~ 1 | sw(f2, f3) | Z1 ~ X1 ", data=data)
    with pytest.raises(NotImplementedError):
        feols(fml="Y  ~ 1 | csw(f2, f3) | Z1 ~ X1 ", data=data)
    # unsupported HC vcov
    with pytest.raises(VcovTypeNotSupportedError):
        feols(fml="Y  ~ 1 | Z1 ~ X1", vcov="HC2", data=data)
    with pytest.raises(VcovTypeNotSupportedError):
        feols(fml="Y  ~ 1 | Z1 ~ X1", vcov="HC3", data=data)


@pytest.mark.skip("Not yet implemented.")
def test_poisson_devpar_count():
    """Check that the dependent variable is a count variable."""
    data = get_data()
    # under determined
    with pytest.raises(AssertionError):
        fepois(fml="Y ~ X1 | X4", data=data)


def test_feols_errors():
    data = pf.get_data()

    with pytest.raises(TypeError):
        pf.feols(fml=1, data=data)

    with pytest.raises(TypeError):
        pf.feols(fml="Y ~ X1", data=1)

    with pytest.raises(TypeError):
        pf.feols(fml="Y ~ X1", data=data, vcov=1)

    with pytest.raises(TypeError):
        pf.feols(fml="Y ~ X1", data=data, fixef_rm=1)

    with pytest.raises(ValueError):
        pf.feols(fml="Y ~ X1", data=data, fixef_rm="f1")

    with pytest.raises(TypeError):
        pf.feols(fml="Y ~ X1", data=data, collin_tol=2)

    with pytest.raises(TypeError):
        pf.feols(fml="Y ~ X1", data=data, collin_tol="2")

    with pytest.raises(ValueError):
        pf.feols(fml="Y ~ X1", data=data, collin_tol=-1.0)

    with pytest.raises(TypeError):
        pf.feols(fml="Y ~ X1", data=data, lean=1)

    with pytest.raises(TypeError):
        pf.feols(fml="Y ~ X1", data=data, fixef_tol="a")

    with pytest.raises(ValueError):
        pf.feols(fml="Y ~ X1", data=data, fixef_tol=1.0)

    with pytest.raises(ValueError):
        pf.feols(fml="Y ~ X1", data=data, fixef_tol=0.0)

    with pytest.raises(ValueError):
        pf.feols(fml="Y ~ X1", data=data, weights_type="qweights", weights="weights")


def test_poisson_errors():
    data = pf.get_data(model="Fepois")
    # iv not supported
    with pytest.raises(NotImplementedError):
        pf.fepois("Y ~ 1 | X1 ~ Z1", data=data)


def test_all_variables_multicollinear():
    data = get_data()
    with pytest.raises(ValueError):
        fit = feols("Y ~ f1 | f1", data=data)  # noqa: F841


def test_wls_errors():
    data = get_data()

    with pytest.raises(AssertionError):
        feols(fml="Y ~ X1", data=data, weights="weights2")

    with pytest.raises(ValueError):
        feols("Y ~ X1", data=data, weights=[1, 2])

    data.loc[0, "weights"] = np.nan
    with pytest.raises(NotImplementedError):
        feols("Y ~ X1", data=data, weights="weights").wildboottest(
            cluster="f1", param="X1", reps=999, seed=12
        )

    # test for ValueError when weights are not positive
    data.loc[10, "weights"] = -1
    with pytest.raises(ValueError):
        feols("Y ~ X1", data=data, weights="weights", vcov="iid")

    # test for ValueError when weights are not numeric
    data["weights"] = data["weights"].astype("str")
    data.loc[10, "weights"] = "a"
    with pytest.raises(ValueError):
        feols("Y ~ X1", data=data, weights="weights", vcov="iid")

    data = get_data()
    with pytest.raises(NotImplementedError):
        feols("Y ~ X1", data=data, weights="weights", vcov="iid").wildboottest(reps=999)


def test_multcomp_errors():
    data = get_data().dropna()

    # param not in model
    fit1 = feols("Y + Y2 ~ X1 | f1", data=data)
    with pytest.raises(ValueError):
        rwolf(fit1.to_list(), param="X2", reps=999, seed=92)


def test_multcomp_sampling_errors():
    data = get_data().dropna()
    # Sampling method not supported in "rwolf"
    fit1 = feols("Y + Y2 ~ X1 | f1", data=data)
    with pytest.raises(ValueError):
        rwolf(fit1.to_list(), param="X1", reps=999, seed=92, sampling_method="abc")


def test_rwolf_error():
    rng = np.random.default_rng(123)

    data = get_data()
    data["f1"] = rng.choice(range(5), len(data), True)
    fit = feols("Y + Y2 ~ X1 | f1", data=data)

    # test for full enumeration warning
    with pytest.warns(UserWarning):
        pf.rwolf(fit.to_list(), "X1", reps=9999, seed=123)


def test_predict_dtype_error():
    data = get_data()
    fit = feols("Y ~ X1 | f1", data=data)

    data["f1"] = data["f1"].fillna(0).astype(int)
    with pytest.warns(UserWarning):
        fit.predict(newdata=data.iloc[0:100])


def test_wildboottest_errors():
    data = get_data()
    fit = feols("Y ~ X1", data=data)
    with pytest.raises(ValueError):
        fit.wildboottest(param="X2", reps=999, seed=213)


def test_summary_errors():
    "Test for appropriate errors when providing objects of type FixestMulti."
    data = get_data()
    fit1 = feols("Y + Y2 ~ X1 | f1", data=data)
    fit2 = feols("Y ~ X1 + X2 | f1", data=data)

    with pytest.raises(TypeError):
        etable([fit1, fit2])
    with pytest.raises(TypeError):
        summary([fit1, fit2])


def test_errors_etable():
    data = get_data()
    fit1 = feols("Y ~ X1", data=data)
    fit2 = feols("Y ~ X1 + X2 | f1", data=data)

    with pytest.raises(AssertionError):
        etable([fit1, fit2], signif_code=[0.01, 0.05])

    with pytest.raises(AssertionError):
        etable([fit1, fit2], signif_code=[0.2, 0.05, 0.1])

    with pytest.raises(AssertionError):
        etable([fit1, fit2], signif_code=[0.1, 0.5, 1.5])

    with pytest.raises(AssertionError):
        etable(
            models=[fit1, fit2],
            custom_stats={
                "conf_int_lb": [
                    fit2._conf_int[0]
                ],  # length of customized statistics not equal to the number of models
                "conf_int_ub": [fit2._conf_int[1]],
            },
            coef_fmt="b se\n[conf_int_lb, conf_int_ub]",
        )

    with pytest.raises(AssertionError):
        etable(
            models=[fit1, fit2],
            custom_stats={
                "conf_int_lb": [
                    [0.1, 0.1, 0.1],
                    fit2._conf_int[0],
                ],  # length of customized statistics not equal to length of model
                "conf_int_ub": [fit1._conf_int[1], fit2._conf_int[1]],
            },
            coef_fmt="b [conf_int_lb, conf_int_ub]",
        )

    with pytest.raises(ValueError):
        etable(
            models=[fit1, fit2],
            custom_stats={
                "b": [
                    fit2._conf_int[0],
                    fit2._conf_int[0],
                ],  # preserved keyword cannot be used as a custom statistic
            },
            coef_fmt="b [se]",
        )


def test_errors_ccv():
    data = get_data().dropna()
    data["D"] = np.random.choice([0, 1], size=len(data))

    # error when D is not binary
    fit = feols("Y ~ X1", data=data, vcov={"CRV1": "f1"})
    with pytest.raises(AssertionError):
        fit.ccv(treatment="X1", pk=0.05, qk=0.5, n_splits=10, seed=929)

    # error when fixed effects in estimation
    fit = feols("Y ~ D | f1", data=data)
    with pytest.raises(NotImplementedError):
        fit.ccv(treatment="D", pk=0.05, qk=0.5, n_splits=10, seed=929)

    # error when treatment not found
    fit = feols("Y ~ D", data=data)
    with pytest.raises(ValueError):
        fit.ccv(treatment="X2", pk=0.05, qk=0.5, n_splits=10, seed=929)

    # error when no cluster variable found
    fit = feols("Y ~ D", data=data)
    with pytest.raises(ValueError):
        fit.ccv(treatment="D", pk=0.05, qk=0.5, n_splits=10, seed=929)

    # error when cluster variable not in data.frame
    with pytest.raises(ValueError):
        fit.ccv(treatment="D", pk=0.05, qk=0.5, n_splits=10, seed=929, cluster="e8")

    # error when two-way clustering
    fit = feols("Y ~ D", data=data, vcov={"CRV1": "f1+f2"})
    with pytest.raises(ValueError):
        fit.ccv(treatment="D", pk=0.05, qk=0.5, n_splits=10, seed=929)

    # error when ccv is attempted on fepois
    pois_data = get_data(model="Fepois").dropna()
    pois_data["D"] = np.random.choice([0, 1], size=len(pois_data))
    fit = fepois("Y ~ D", data=pois_data)
    with pytest.raises(AssertionError):
        fit.ccv(treatment="D", pk=0.05, qk=0.5, n_splits=10, seed=929)

    # same for IV
    fit = feols("Y ~ 1 | D ~ Z1", data=data)
    with pytest.raises(AssertionError):
        fit.ccv(treatment="D", pk=0.05, qk=0.5, n_splits=10, seed=929)


def test_errors_confint():
    data = get_data()
    fit = feols("Y ~ X1", data=data)
    with pytest.raises(ValueError):
        fit.confint(alpha=0.5, keep=["abababa"])


def test_i_error():
    data = get_data()
    data["f2"] = pd.Categorical(data["f2"])

    with pytest.raises(ValueError):
        feols("Y ~ i(f1, f2)", data)

    data["f2"] = data["f2"].astype("object")
    with pytest.raises(ValueError):
        feols("Y ~ i(f1, f2)", data)

    with pytest.raises(FactorEvaluationError):
        feols("Y ~ i(f1, X1, ref=a)", data)


def test_plot_error():
    df = get_data()
    fit = feols("Y ~ X1", data=df)
    with pytest.raises(
        ValueError, match="plot_backend must be either 'lets_plot' or 'matplotlib'."
    ):
        fit.coefplot(plot_backend="plotnine")

    fit_multi = feols("Y + Y2 ~ i(f1)", data=df)
    with pytest.raises(ValueError):
        fit_multi.coefplot(joint=True)

    with pytest.raises(ValueError):
        fit_multi.iplot(joint="both")


def test_ritest_error(data):
    fit = pf.feols("Y ~ X1", data=data)

    with pytest.raises(ValueError, match="X2 not found in the model's coefficients."):
        fit.ritest(resampvar="X2", reps=1000)

    with pytest.raises(ValueError, match="CLUST is not found in the data"):
        fit.ritest(resampvar="X1", reps=1000, cluster="CLUST")

    with pytest.raises(
        ValueError, match="type must be 'randomization-t' or 'randomization-c."
    ):
        fit.ritest(resampvar="X1", reps=1000, type="a")

    with pytest.raises(AssertionError):
        fit.ritest(resampvar="X1", reps=100.4)

    with pytest.raises(ValueError):
        fit.ritest(resampvar="X1", cluster="f1", reps=100)

    with pytest.raises(NotImplementedError):
        fit_iv = pf.feols("Y ~ 1 | X1 ~ Z1", data=data)
        fit_iv.ritest(resampvar="X1", reps=100)

    with pytest.raises(NotImplementedError):
        fit_wls = pf.feols("Y ~ X1", data=data, weights="weights")
        fit_wls.ritest(resampvar="X1", reps=100)

    with pytest.raises(ValueError):
        "Incorrect plot backend."
        fit.ritest(resampvar="X1", reps=100, store_ritest_statistics=True)
        fit.plot_ritest(plot_backend="a")

    with pytest.raises(ValueError):
        "No test_statistics found in the model."
        fit = pf.feols("Y ~ X1", data=data)
        fit.ritest(resampvar="X1", reps=100)
        fit.plot_ritest()


def test_wald_test_invalid_distribution():
    data = pd.read_csv("pyfixest/did/data/df_het.csv")
    data = data.iloc[1:3000]

    fml = "dep_var ~ treat"
    fit = feols(fml, data, vcov={"CRV1": "year"}, ssc=ssc(adj=False))

    with pytest.raises(ValueError):
        fit.wald_test(R=np.array([[1, -1]]), distribution="abc")


def test_wald_test_R_q_column_consistency():
    data = pd.read_csv("pyfixest/did/data/df_het.csv")
    data = data.iloc[1:3000]
    fml = "dep_var ~ treat"
    fit = feols(fml, data, vcov={"CRV1": "year"}, ssc=ssc(adj=False))

    # Test with R.size[1] == number of coeffcients
    with pytest.raises(ValueError):
        fit.wald_test(R=np.array([[1, 0, 0]]))

    # Test with q type
    with pytest.raises(ValueError):
        fit.wald_test(R=np.array([[1, 0]]), q="invalid type q")

    # Test with q being a one-dimensional array or a scalar.
    with pytest.raises(ValueError):
        fit.wald_test(R=np.array([[1, 0], [0, 1]]), q=np.array([[0, 1]]))

    # q must have the same number of rows as R
    with pytest.raises(ValueError):
        fit.wald_test(
            R=np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]), q=np.array([[0, 1]])
        )


def setup_feiv_instance():
    # Setup necessary data for Feiv

    data = pf.get_data()
    return pf.feols("Y ~ 1 | X1 ~ Z1", data=data)


def test_IV_first_stage_invalid_model_type():
    class NotFeols:
        # Dummy class for testing invalid model type
        pass

    invalid_model = NotFeols()

    with pytest.raises(TypeError):
        feiv_instance = setup_feiv_instance()
        feiv_instance.first_stage(invalid_model)  # This should raise TypeError


def test_IV_Diag_unsupported_statistics():
    feiv_instance = setup_feiv_instance()

    unsupported_statistics = ["unsupported_stat"]

    with pytest.raises(ValueError):
        feiv_instance.IV_Diag(statistics=unsupported_statistics)


def test_errors_compressed():
    data = pf.get_data()

    # no more than two fixed effects
    with pytest.raises(NotImplementedError):
        pf.feols("Y ~ X1 | f1 + f2 + f3", data=data, use_compression=True)

    # cluster variables not in model
    with pytest.raises(NotImplementedError):
        pf.feols("Y ~ X1", vcov={"CRV1": "f1"}, data=data, use_compression=True)

    with pytest.raises(NotImplementedError):
        pf.feols("Y ~ C(f1)", vcov={"CRV1": "f1"}, data=data, use_compression=True)

    with pytest.raises(NotImplementedError):
        pf.feols("Y ~ X1 | f1", data=data, use_compression=True, vcov={"CRV1": "f1+f2"})

    # only CVR supported for Mundlak
    with pytest.raises(NotImplementedError):
        pf.feols("Y ~ X1 | f1", data=data, use_compression=True, vcov="iid")

    # crv3 inference:
    with pytest.raises(NotImplementedError):
        pf.feols("Y ~ X1 | f1", vcov={"CRV3": "f1"}, data=data, use_compression=True)

    # prediction:
    with pytest.raises(NotImplementedError):
        pf.feols("Y ~ X1 | f1", data=data, use_compression=True).predict()

    # argument errors
    with pytest.raises(TypeError):
        pf.feols("Y ~ X1", data=data, use_compression=True, vcov="iid", reps=1.2)

    with pytest.raises(ValueError):
        pf.feols("Y ~ X1", data=data, use_compression=True, vcov="iid", reps=-1)

    with pytest.raises(TypeError):
        pf.feols("Y ~ X1", data=data, use_compression=True, vcov="iid", seed=1.2)

    # no support for IV
    with pytest.raises(NotImplementedError):
        pf.feols("Y ~ 1 | X1 ~ Z1", data=data, use_compression=True)

    # no support for WLS
    with pytest.raises(NotImplementedError):
        pf.feols("Y ~ X1", data=data, weights="weights", use_compression=True)


def test_errors_panelview():
    """Test all ValueError conditions in panelview."""
    sample_df = pd.DataFrame(
        {
            "unit": [1, 2, 3, 4],
            "year": [2001, 2002, 2003, 2004],
            "treat": [0, 1, 1, 0],
            "dep_var": [10, 20, 30, 40],
        }
    )

    # 1. Test for missing 'unit' column
    with pytest.raises(ValueError, match="Column 'unit' not found in data."):
        pf.panelview(sample_df.drop(columns=["unit"]), "unit", "year", "treat")

    # 2. Test for missing 'year' column (time)
    with pytest.raises(ValueError, match="Column 'year' not found in data."):
        pf.panelview(sample_df.drop(columns=["year"]), "unit", "year", "treat")

    # 3. Test for missing 'treat' column
    with pytest.raises(ValueError, match="Column 'treat' not found in data."):
        pf.panelview(sample_df.drop(columns=["treat"]), "unit", "year", "treat")

    # 4. Test for missing outcome column
    with pytest.raises(
        ValueError, match="Outcome column 'nonexistent_col' not found in data."
    ):
        pf.panelview(sample_df, "unit", "year", "treat", outcome="nonexistent_col")

    # 5. Test for 'collapse_to_cohort' and 'subsamp' used together
    with pytest.raises(
        ValueError,
        match="Cannot use 'collapse_to_cohort' together with 'subsamp' or 'units_to_plot'.",
    ):
        pf.panelview(
            sample_df, "unit", "year", "treat", collapse_to_cohort=True, subsamp=10
        )

    # 6. Test for 'collapse_to_cohort' and 'units_to_plot' used together
    with pytest.raises(
        ValueError,
        match="Cannot use 'collapse_to_cohort' together with 'subsamp' or 'units_to_plot'.",
    ):
        pf.panelview(
            sample_df,
            "unit",
            "year",
            "treat",
            collapse_to_cohort=True,
            units_to_plot=[1, 2],
        )


@pytest.mark.parametrize(
    "split, fsplit, expected_exception, error_message",
    [
        # Test TypeError for non-string 'split'
        (123, None, TypeError, "The function argument split needs to be of type str."),
        # Test TypeError for non-string 'fsplit'
        (None, 456, TypeError, "The function argument fsplit needs to be of type str."),
        # Test ValueError for split and fsplit not being identical
        (
            "split_column",
            "different_column",
            ValueError,
            r"Arguments split and fsplit are both specified, but not identical",
        ),
        # Test KeyError for invalid 'split' column
        (
            "invalid_column",
            None,
            KeyError,
            "Column 'invalid_column' not found in data.",
        ),
        # Test KeyError for invalid 'fsplit' column
        (
            None,
            "invalid_column",
            KeyError,
            "Column 'invalid_column' not found in data.",
        ),
    ],
)
def test_split_fsplit_errors(data, split, fsplit, expected_exception, error_message):
    with pytest.raises(expected_exception, match=error_message):
        pf.feols("Y~X1", data=data, split=split, fsplit=fsplit)


def test_separation_check_validations():
    data = pd.DataFrame(
        {
            "Y": [1, 2, 3],
            "X1": [4, 5, 6],
        }
    )

    with pytest.raises(
        ValueError,
        match="The function argument `separation_check` must be a list of strings containing 'fe' and/or 'ir'.",
    ):
        pf.fepois("Y ~ X1", data=data, separation_check=["a"])

    with pytest.raises(
        TypeError,
        match="The function argument `separation_check` must be of type list.",
    ):
        pf.fepois("Y ~ X1", data=data, separation_check="fe")

    with pytest.raises(
        ValueError,
        match="The function argument `separation_check` must be a list of strings containing 'fe' and/or 'ir'.",
    ):
        pf.fepois("Y ~ X1", data=data, separation_check=["fe", "invalid"])


def test_gelbach_errors():
    rng = np.random.default_rng(123)

    data = gelbach_data(nobs=100)
    data["f1"] = rng.choice(range(5), len(data), True)
    data["weights"] = rng.uniform(0.5, 1.5, len(data))

    fit = pf.feols("y ~ x1 + x21 + x22 + x23", data=data)

    with pytest.raises(ValueError, match=r"x32 is not in the mediator names."):
        fit.decompose(param="x1", combine_covariates={"g1": ["x32"]})

    with pytest.raises(ValueError, match=r"{'x21'} is in both g1 and g2."):
        fit.decompose(param="x1", combine_covariates={"g1": ["x21"], "g2": ["x21"]})

    # error with IV
    with pytest.raises(NotImplementedError):
        pf.feols("y ~ 1 | x1 ~ x21", data=data).decompose(
            param="x1", combine_covariates={"g1": ["x21"]}
        )

    # error with WLS
    with pytest.raises(NotImplementedError):
        pf.feols("y ~ x1", data=data, weights="weights").decompose(
            param="x1", combine_covariates={"g1": ["x21"]}
        )

    # error with Poisson
    with pytest.raises(NotImplementedError):
        dt = pf.get_data(model="Fepois")
        pf.fepois("Y ~ X1", data=dt).decompose(
            param="X1", combine_covariates={"g1": ["x21"]}
        )


def test_glm_errors():
    "Test that dependent variable must be binary for probit and logit models."
    data = pf.get_data()
    with pytest.raises(
        ValueError, match="The dependent variable must have two unique values."
    ):
        pf.feglm("Y ~ X1", data=data, family="probit")
    with pytest.raises(
        ValueError, match="The dependent variable must have two unique values."
    ):
        pf.feglm("Y ~ X1", data=data, family="logit")

    data["Y"] = np.where(data["Y"] > 0, 2, 0)
    with pytest.raises(
        ValueError, match=r"The dependent variable must be binary \(0 or 1\)."
    ):
        pf.feglm("Y ~ X1", data=data, family="probit")
    with pytest.raises(
        ValueError, match=r"The dependent variable must be binary \(0 or 1\)."
    ):
        pf.feglm("Y ~ X1", data=data, family="logit")

    data["Y"] = np.where(data["Y"] > 0, 1, 0)
    with pytest.raises(
        NotImplementedError, match=r"Fixed effects are not yet supported for GLMs."
    ):
        pf.feglm("Y ~ X1 | f1", data=data, family="probit")


def test_prediction_errors_glm():
    "Test that the prediction errors not supported for GLM models."
    data = pf.get_data()
    data["Y"] = np.where(data["Y"] > 0, 1, 0)

    fit_gaussian = pf.feglm("Y ~ X1", data=data, family="gaussian")
    fit_probit = pf.feglm("Y ~ X1", data=data, family="probit")
    fit_logit = pf.feglm("Y ~ X1", data=data, family="logit")
    fit_pois = pf.fepois("Y ~ X1", data=data)

    for model in [fit_gaussian, fit_probit, fit_logit, fit_pois]:
        with pytest.raises(
            NotImplementedError, match="Prediction with standard errors"
        ):
            model.predict(se_fit=True)


def test_empty_vcov_error():
    data = pf.get_data()
    fit = pf.feols("Y ~ 1 | f1", data=data)

    with pytest.warns(UserWarning):
        fit.tidy()


def test_errors_quantreg(data):
    data = data.dropna()

    # error for CRV3
    with pytest.raises(VcovTypeNotSupportedError):
        pf.quantreg("Y ~ X1", data=data, vcov={"CRV3": "f1"})

    # error for two-way clustering
    with pytest.raises(NotImplementedError):
        pf.quantreg("Y ~ X1", data=data, vcov={"CRV1": "f1+f2"})

    # error for quantile outside [0, 1]
    with pytest.raises(ValueError, match="quantile must be between 0 and 1"):
        pf.quantreg("Y ~ X1", data=data, quantile=1.1)
    with pytest.raises(ValueError, match="quantile must be between 0 and 1"):
        pf.quantreg("Y ~ X1", data=data, quantile=-0.1)

    # error if list provided and quantile is 0, 1, or outside
    with pytest.raises(ValueError, match="quantile must be a list of floats"):
        pf.quantreg("Y ~ X1", data=data, quantile=["0.1", "0.2"])

    # error when fixed effects in formula
    with pytest.raises(NotImplementedError):
        pf.quantreg("Y ~ X1 | f1", data=data)

    # error for invalid method
    with pytest.raises(ValueError, match="`method` must be one of {fn, pfn}"):
        pf.quantreg("Y ~ X1", data=data, method="invalid_method")

    # error for invalid tolerance
    @pytest.mark.parametrize("tol", [0, 1, -0.1, 1.1])
    def test_invalid_tolerance(tol):
        with pytest.raises(ValueError, match=r"tol must be in \(0, 1\)"):
            pf.quantreg("Y ~ X1", data=data, tol=tol)



class FixedEffectInteractionError(Exception):  # noqa: D101
    pass


class CovariateInteractionError(Exception):  # noqa: D101
    pass


class DuplicateKeyError(Exception):  # noqa: D101
    pass


class EndogVarsAsCovarsError(Exception):  # noqa: D101
    pass


class InstrumentsAsCovarsError(Exception):  # noqa: D101
    pass


class UnderDeterminedIVError(Exception):  # noqa: D101
    pass


class UnsupportedMultipleEstimationSyntax(Exception):  # noqa: D101
    pass


class VcovTypeNotSupportedError(Exception):  # noqa: D101
    pass


class NanInClusterVarError(Exception):  # noqa: D101
    pass


class DepvarIsNotNumericError(Exception):  # noqa: D101
    pass


class NonConvergenceError(Exception):  # noqa: D101
    pass


class MatrixNotFullRankError(Exception):  # noqa: D101
    pass


class EmptyDesignMatrixError(Exception):  # noqa: D101
    pass


class FeatureDeprecationError(Exception):  # noqa: D101
    pass


class EmptyVcovError(Exception):  # noqa: D101
    pass


__all__ = [
    "CovariateInteractionError",
    "DepvarIsNotNumericError",
    "DuplicateKeyError",
    "EmptyDesignMatrixError",
    "EmptyVcovError",
    "EndogVarsAsCovarsError",
    "FeatureDeprecationError",
    "FixedEffectInteractionError",
    "InstrumentsAsCovarsError",
    "MatrixNotFullRankError",
    "NanInClusterVarError",
    "NonConvergenceError",
    "UnderDeterminedIVError",
    "UnsupportedMultipleEstimationSyntax",
    "VcovTypeNotSupportedError",
]



from pyfixest.estimation.FormulaParser import (
    _dict_to_list_of_formulas,
    _find_multiple_estimation_syntax,
    _input_formula_to_dict,
)


def test_input_formula_to_dict():
    assert _input_formula_to_dict("a+b+c") == {"constant": ["a", "b", "c"]}

    assert _input_formula_to_dict("sw(x,y)") == {"constant": [], "sw": ["x", "y"]}

    assert _input_formula_to_dict("a+sw0(x,y)+d") == {
        "constant": ["a", "d"],
        "sw0": ["x", "y"],
    }

    assert _input_formula_to_dict("csw(x,y)") == {"constant": [], "csw": ["x", "y"]}

    assert _input_formula_to_dict("csw0(x,y,z)") == {
        "constant": [],
        "csw0": ["x", "y", "z"],
    }

    assert _input_formula_to_dict("a+b+csw0(x,y,z)") == {
        "constant": ["a", "b"],
        "csw0": ["x", "y", "z"],
    }


def test_dict_to_list_of_formulas():
    # assert _dict_to_list_of_formulas([]) == []

    assert _dict_to_list_of_formulas({"constant": ["x", "y"], "sw0": ["a", "b"]}) == [
        "x+y",
        "x+y+a",
        "x+y+b",
    ]
    assert _dict_to_list_of_formulas({"constant": [], "sw0": ["a", "b"]}) == [
        "0",
        "a",
        "b",
    ]
    assert _dict_to_list_of_formulas({"constant": ["x", "y"], "sw0": []}) == ["x+y"]

    assert _dict_to_list_of_formulas({"constant": ["x", "y"], "sw": ["a", "b"]}) == [
        "x+y+a",
        "x+y+b",
    ]
    assert _dict_to_list_of_formulas({"constant": [], "sw": ["a", "b"]}) == ["a", "b"]
    assert _dict_to_list_of_formulas({"constant": ["x", "y"], "sw": []}) == ["x+y"]

    assert _dict_to_list_of_formulas({"constant": ["x", "y"], "csw0": ["a", "b"]}) == [
        "x+y",
        "x+y+a",
        "x+y+a+b",
    ]
    assert _dict_to_list_of_formulas({"constant": [], "csw0": ["a", "b"]}) == [
        "0",
        "a",
        "a+b",
    ]
    assert _dict_to_list_of_formulas({"constant": ["x", "y"], "csw0": []}) == ["x+y"]

    assert _dict_to_list_of_formulas({"constant": ["x", "y"], "csw": ["a", "b"]}) == [
        "x+y+a",
        "x+y+a+b",
    ]
    assert _dict_to_list_of_formulas({"constant": [], "csw": ["a", "b"]}) == [
        "a",
        "a+b",
    ]
    assert _dict_to_list_of_formulas({"constant": ["x", "y"], "csw": []}) == ["x+y"]


def test_find_multiple_estimation_syntax_no_match():
    x = "a + b + c"
    assert _find_multiple_estimation_syntax(x) == (x, None)


def test_find_multiple_estimation_syntax_sw():
    x = "sw(a, b, c)"
    expected = (["a", " b", " c"], "sw")
    assert _find_multiple_estimation_syntax(x) == expected


def test_find_multiple_estimation_syntax_csw():
    x = "csw(a, b, c)"
    expected = (["a", " b", " c"], "csw")
    assert _find_multiple_estimation_syntax(x) == expected


def test_find_multiple_estimation_syntax_sw0():
    x = "sw0(a, b, c)"
    expected = (["a", " b", " c"], "sw0")
    assert _find_multiple_estimation_syntax(x) == expected


def test_find_multiple_estimation_syntax_csw0():
    x = "csw0(a, b, c)"
    expected = (["a", " b", " c"], "csw0")
    assert _find_multiple_estimation_syntax(x) == expected



import numpy as np
import pandas as pd
import pytest

import pyfixest as pf
from pyfixest.did.estimation import did2s, event_study


@pytest.fixture
def data():
    df_het = pd.read_csv("pyfixest/did/data/df_het.csv")
    return df_het


def test_event_study_twfe(data):
    twfe = event_study(
        data=data,
        yname="dep_var",
        idname="state",
        tname="year",
        gname="g",
        att=True,
        estimator="twfe",
    )

    twfe_feols = pf.feols("dep_var ~ treat | state + year", data=data)

    assert np.allclose(twfe.coef().values, twfe_feols.coef().values), (
        "TWFE coefficients are not the same."
    )
    assert np.allclose(twfe.se().values, twfe_feols.se().values), (
        "TWFE standard errors are not the same."
    )
    assert np.allclose(twfe.pvalue().values, twfe_feols.pvalue().values), (
        "TWFE p-values are not the same."
    )

    # TODO - minor difference, likely due to how z statistic is
    # calculated

    # assert np.allclose(
    #    twfe.confint().values, twfe_feols.confint().values
    # ), "TWFE confidence intervals are not the same."


def test_event_study_did2s(data):
    event_study_did2s = event_study(
        data=data,
        yname="dep_var",
        idname="state",
        tname="year",
        gname="g",
        att=True,
        estimator="did2s",
    )

    fit_did2s = did2s(
        data=data,
        yname="dep_var",
        first_stage="~ 0 | state + year",
        second_stage="~treat",
        treatment="treat",
        cluster="state",
    )

    assert np.allclose(event_study_did2s.coef().values, fit_did2s.coef().values), (
        "DID2S coefficients are not the same."
    )
    assert np.allclose(event_study_did2s.se().values, fit_did2s.se().values), (
        "DID2S standard errors are not the same."
    )
    assert np.allclose(event_study_did2s.pvalue().values, fit_did2s.pvalue().values), (
        "DID2S p-values are not the same."
    )
    assert np.allclose(
        event_study_did2s.confint().values, fit_did2s.confint().values
    ), "DID2S confidence intervals are not the same."


# ---------------------------------------------------------------------------------
# test errors


# Test case for 'data' must be a pandas DataFrame
def test_event_study_invalid_data_type(data):
    with pytest.raises(AssertionError, match="data must be a pandas DataFrame"):
        event_study(
            data="invalid_data",  # Invalid data type, should be pd.DataFrame
            yname="dep_var",
            idname="state",
            tname="year",
            gname="g",
            estimator="twfe",
        )


# Test case for 'yname' must be a string
def test_event_study_invalid_yname_type(data):
    with pytest.raises(AssertionError, match="yname must be a string"):
        event_study(
            data=data,
            yname=123,  # Invalid yname type, should be str
            idname="state",
            tname="year",
            gname="g",
            estimator="twfe",
        )


# Test case for 'idname' must be a string
def test_event_study_invalid_idname_type(data):
    with pytest.raises(AssertionError, match="idname must be a string"):
        event_study(
            data=data,
            yname="dep_var",
            idname=123,  # Invalid idname type, should be str
            tname="year",
            gname="g",
            estimator="twfe",
        )


# Test case for 'tname' must be a string
def test_event_study_invalid_tname_type(data):
    with pytest.raises(AssertionError, match="tname must be a string"):
        event_study(
            data=data,
            yname="dep_var",
            idname="state",
            tname=2020,  # Invalid tname type, should be str
            gname="g",
            estimator="twfe",
        )


# Test case for 'gname' must be a string
def test_event_study_invalid_gname_type(data):
    with pytest.raises(AssertionError, match="gname must be a string"):
        event_study(
            data=data,
            yname="dep_var",
            idname="state",
            tname="year",
            gname=2020,  # Invalid gname type, should be str
            estimator="twfe",
        )


# Test case for 'xfml' must be a string or None
def test_event_study_invalid_xfml_type(data):
    with pytest.raises(AssertionError, match="xfml must be a string or None"):
        event_study(
            data=data,
            yname="dep_var",
            idname="state",
            tname="year",
            gname="g",
            xfml=123,  # Invalid xfml type, should be str or None
            estimator="twfe",
        )


# Test case for 'estimator' must be a string
def test_event_study_invalid_estimator_type(data):
    with pytest.raises(AssertionError, match="estimator must be a string"):
        event_study(
            data=data,
            yname="dep_var",
            idname="state",
            tname="year",
            gname="g",
            estimator=123,  # Invalid estimator type, should be str
        )


# Test case for 'att' must be a boolean
def test_event_study_invalid_att_type(data):
    with pytest.raises(AssertionError, match="att must be a boolean"):
        event_study(
            data=data,
            yname="dep_var",
            idname="state",
            tname="year",
            gname="g",
            att="True",  # Invalid att type, should be bool
            estimator="twfe",
        )


# Test case for 'cluster' must be a string
def test_event_study_invalid_cluster_type(data):
    with pytest.raises(AssertionError, match="cluster must be a string"):
        event_study(
            data=data,
            yname="dep_var",
            idname="state",
            tname="year",
            gname="g",
            estimator="twfe",
            cluster=123,  # Invalid cluster type, should be str
        )


# Test case for unsupported estimator (triggering NotImplementedError)
def test_event_study_unsupported_estimator(data):
    with pytest.raises(NotImplementedError, match="Estimator not supported"):
        event_study(
            data=data,
            yname="dep_var",
            idname="state",
            tname="year",
            gname="g",
            estimator="unsupported",  # Unsupported estimator
        )



# ----------------------------------------------------------------------
#
# SPDX-License-Identifier: MIT
#
# Copyright (c) 2025  PyFixest Authors
#
# This function is a Python/Numba re-implementation of the algorithm
# published by Laurent Berge in the **fixest**
# project (see See the fixest repo
# [here](https:#github.com/lrberge/fixest/blob/a4d1a9bea20aa7ab7ab0e0f1d2047d8097971ad7/src/lm_related.cpp#L130)),
# originally licensed under GPL-3.0.  Laurent Berge granted the maintainers of this
# project an irrevocable, written permission to
# redistribute and re-license the relevant code under the MIT License.
# The full text of that permission is archived at:
#
# docs/THIRD_PARTY_PERMISSIONS.md
# ----------------------------------------------------------------------

import numba as nb
import numpy as np


@nb.njit(parallel=False)
def _find_collinear_variables_nb(
    X: np.ndarray, tol: float = 1e-10
) -> tuple[np.ndarray, int, bool]:
    """
    Detect multicollinear variables.

    Detect multicollinear variables, replicating Laurent Berge's C++ implementation
    from the fixest package. See the fixest repo [here](https://github.com/lrberge/fixest/blob/a4d1a9bea20aa7ab7ab0e0f1d2047d8097971ad7/src/lm_related.cpp#L130)

    Parameters
    ----------
    X : numpy.ndarray
        A symmetric matrix X used to check for multicollinearity.
    tol : float
        The tolerance level for the multicollinearity check.

    Returns
    -------
    - id_excl (numpy.ndarray): A boolean array, where True indicates a collinear
        variable.
    - n_excl (int): The number of collinear variables.
    - all_removed (bool): True if all variables are identified as collinear.
    """
    K = X.shape[1]
    R = np.zeros((K, K))
    id_excl = np.zeros(K, dtype=np.int32)
    n_excl = 0
    min_norm = X[0, 0]

    for j in range(K):
        R_jj = X[j, j]
        for k in range(j):
            if id_excl[k]:
                continue
            R_jj -= R[k, j] * R[k, j]

        if R_jj < tol:
            n_excl += 1
            id_excl[j] = 1

            if n_excl == K:
                all_removed = True
                return id_excl.astype(np.bool_), n_excl, all_removed

            continue

        if min_norm > R_jj:
            min_norm = R_jj

        R_jj = np.sqrt(R_jj)
        R[j, j] = R_jj

        for i in range(j + 1, K):
            value = X[i, j]
            for k in range(j):
                if id_excl[k]:
                    continue
                value -= R[k, i] * R[k, j]
            R[j, i] = value / R_jj

    return id_excl.astype(np.bool_), n_excl, False



import sys

import numpy as np
import pandas as pd
import pytest

from pyfixest.estimation.estimation import feols
from pyfixest.utils.utils import get_data


def test_confint():
    """Test the confint method of the feols class."""
    data = get_data()
    fit = feols("Y ~ X1 + X2 + C(f1)", data=data)
    confint = fit.confint()

    np.testing.assert_allclose(confint, fit.confint(alpha=0.05))
    assert np.all(confint.loc[:, "2.5%"] == fit.confint(alpha=0.05).loc[:, "2.5%"])
    assert np.all(confint.loc[:, "97.5%"] == fit.confint(alpha=0.05).loc[:, "97.5%"])
    assert np.all(confint.loc[:, "2.5%"] < fit.confint(alpha=0.10).loc[:, "5.0%"])
    assert np.all(confint.loc[:, "97.5%"] > fit.confint(alpha=0.10).loc[:, "95.0%"])

    # test keep, drop, and exact_match
    assert fit.confint(keep="X1", exact_match=True).shape[0] == 1
    assert (
        fit.confint(drop=["X2"], exact_match=True).shape[0] == len(fit._coefnames) - 1
    )
    assert fit.confint(keep="X").shape[0] == 2

    # simultaneous CIs: simultaneous CIs always wider
    for _ in range(5):
        assert np.all(
            confint.loc[:, "2.5%"] > fit.confint(alpha=0.05, joint=True).loc[:, "2.5%"]
        )
        assert np.all(
            confint.loc[:, "97.5%"]
            < fit.confint(alpha=0.05, joint=True).loc[:, "97.5%"]
        )

    # test seeds
    confint1 = fit.confint(joint=True, seed=1)
    confint2 = fit.confint(joint=True, seed=1)
    confint3 = fit.confint(joint=True, seed=2)

    assert np.all(confint1 == confint2)
    assert np.all(confint1 != confint3)


@pytest.mark.skipif(sys.version_info >= (3, 12), reason="requires python3.11 or lower.")
def test_against_doubleml():
    """Test joint CIs against DoubleML."""
    import doubleml as dml
    from sklearn.base import clone
    from sklearn.linear_model import LinearRegression

    rng = np.random.default_rng(2002)
    n_obs = 5_000
    n_vars = 100
    X = rng.normal(size=(n_obs, n_vars))
    theta = np.array([3.0, 3.0, 3.0])
    y = np.dot(X[:, :3], theta) + rng.standard_normal(size=(n_obs,))

    dml_data = dml.DoubleMLData.from_arrays(X[:, 10:], y, X[:, :10])
    learner = LinearRegression()
    ml_l = clone(learner)
    ml_m = clone(learner)
    dml_plr = dml.DoubleMLPLR(dml_data, ml_l, ml_m)
    dml_res = dml_plr.fit().bootstrap(n_rep_boot=10_000).confint(joint=True)

    df = pd.DataFrame(
        np.c_[y, X], columns=["y"] + ["X_" + str(x) for x in range(n_vars)]
    )
    m = feols(
        f"y ~ -1 + {'+'.join(['X_' + str(x) for x in range(n_vars)])}",
        df,
        vcov="hetero",
    )
    pyfixest_res = m.confint(keep="X_.$", reps=10_000, joint=True)

    assert np.all(np.abs(dml_res.values - pyfixest_res.values) < 1e-2)



from typing import Optional, cast

import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from scipy.sparse.linalg import spsolve

from pyfixest.did.did import DID
from pyfixest.estimation.estimation import feols
from pyfixest.estimation.feols_ import Feols
from pyfixest.estimation.FormulaParser import FixestFormulaParser
from pyfixest.estimation.model_matrix_fixest_ import model_matrix_fixest


class DID2S(DID):
    """
    Difference-in-Differences estimation using Gardner(2021) two-step DID2S estimator.

    Multiple parts of this code are direct translations of Kyle Butt's R code
    `did2s`, published under MIT license: https://github.com/kylebutts/did2s/tree/main.

    Attributes
    ----------
    data : pandas.DataFrame
        The DataFrame containing all variables.
    yname : str
        The name of the dependent variable.
    idname : str
        The name of the identifier variable.
    tname : str
        Variable name for calendar period. Must be an integer in the format
        YYYYMMDDHHMMSS, i.e. it must be possible to compare two dates via '>'.
        Datetime variables are currently not accepted.
    gname : str
        unit-specific time of initial treatment. Must be an integer in the format
        YYYYMMDDHHMMSS, i.e. it must be possible to compare two dates via '>'.
        Datetime variables are currently not accepted. Never treated units
        must have a value of 0.
    xfml : str
        The formula for the covariates.
    att : str
        Whether to estimate the pooled average treatment effect on the treated
        (ATT) or the canonical event study design with all leads and lags / the
        ATT for each period. Default is True.
    cluster : str
        The name of the cluster variable.
    weights : Optional[str].
        Default is None. Weights for WLS estimation. If None, all observations
        are weighted equally. If a string, the name of the column in `data` that
        contains the weights.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        yname: str,
        idname: str,
        tname: str,
        gname: str,
        cluster: str,
        weights: Optional[str] = None,
        att: bool = True,
        xfml: Optional[str] = None,
    ):
        super().__init__(
            data=data,
            yname=yname,
            idname=idname,
            tname=tname,
            gname=gname,
            xfml=xfml,
            att=att,
            cluster=cluster,
        )

        self._estimator = "did2s"

        if self._xfml is not None:
            self._fml1 = f" ~ {xfml} | {idname} + {tname}"
            self._fml2 = f" ~ 0 + is_treated + {xfml}"
        else:
            self._fml1 = f" ~ 0 | {idname} + {tname}"
            self._fml2 = " ~ 0 + is_treated"

        # first and second stage residuals
        self._first_u = np.array([])
        self._second_u = np.array([])

        # column name with weights. None by default
        self._weights_name = weights

    def estimate(self):
        """Estimate the two-step DID2S model."""
        return _did2s_estimate(
            data=self._data,
            yname=self._yname,
            _first_stage=self._fml1,
            _second_stage=self._fml2,
            weights=self._weights_name,
            treatment="is_treated",
        )  # returns triple Feols, first_u, second_u

    def vcov(self):
        """
        Variance-covariance matrix.

        Calculates the variance-covariance matrix of the coefficient estimates
        for the Difference-in-Differences (DiD) estimator.

        Returns
        -------
        array_like
            The variance-covariance matrix of the coefficient estimates.
        """
        return _did2s_vcov(
            data=self._data,
            yname=self._yname,
            first_stage=self._fml1,
            second_stage=self._fml2,
            treatment="is_treated",
            first_u=self._first_u,
            second_u=self._second_u,
            cluster=self._cluster,
            weights=self._weights_name,
        )

    def iplot(
        self,
        alpha: float = 0.05,
        figsize: tuple[int, int] = (500, 300),
        yintercept: Optional[int] = None,
        xintercept: Optional[int] = None,
        rotate_xticks: int = 0,
        title: str = "DID2S Event Study Estimate",
        coord_flip: bool = False,
    ):
        """Plot DID estimates."""
        self.iplot(
            alpha=alpha,
            figsize=figsize,
            yintercept=yintercept,
            xintercept=xintercept,
            rotate_xticks=rotate_xticks,
            title=title,
            coord_flip=coord_flip,
        )

    def tidy(self):  # noqa: D102
        return self.tidy()

    def summary(self):  # noqa: D102
        return self.summary()


def _did2s_estimate(
    data: pd.DataFrame,
    yname: str,
    _first_stage: str,
    _second_stage: str,
    treatment: str,
    weights: Optional[str] = None,
):
    """
    Estimate the two-step DID2S model.

    Parameters
    ----------
    data: pd.DataFrame
        The DataFrame containing all variables.
    yname:str
        The name of the dependent variable.
    _first_stage: str
        The formula for the first stage.
    _second_stage: str
        The formula for the second stage.
    treatment: str
        The name of the treatment variable. Must be boolean.
    weights : Optional[str].
        Default is None. Weights for WLS estimation. If None, all observations
        are weighted equally. If a string, the name of the column in `data` that
        contains the weights.

    Returns
    -------
    object
        A fitted model object of class feols and the first and second stage residuals.
    """
    _first_stage_full = f"{yname} {_first_stage}"
    _second_stage_full = f"{yname}_hat {_second_stage}"

    if treatment is not None:
        if treatment not in data.columns:
            raise ValueError(f"The variable {treatment} is not in the data.")
        treat_u = data[treatment].unique()
        if len(treat_u) != 2:
            raise ValueError(
                f"The treatment variable {treatment} must have 2 unique values but it has unique values {treat_u}."
            )
        if data[treatment].dtype in ["bool", "int64", "int32", "float64", "float32"]:
            data[treatment] = data[treatment].astype(bool)
        else:
            raise ValueError(
                f"The treatment variable {treatment} must be boolean or numeric but it is of type {data[treatment].dtype}."
            )
        _not_yet_treated_data = data[data[treatment] == False]  # noqa: E712
    else:
        _not_yet_treated_data = data[data["is_treated"] == False]  # noqa: E712

    # check if first stage formulas has fixed effects
    if "|" not in _first_stage:
        raise ValueError("First stage formula must contain fixed effects.")
    # check if second stage formulas has fixed effects
    if "|" in _second_stage:
        raise ValueError("Second stage formula must not contain fixed effects.")

    # estimate first stage
    fit1 = cast(
        Feols,
        feols(
            fml=_first_stage_full,
            data=_not_yet_treated_data,
            vcov="iid",
            weights=weights,
        ),
    )  # iid as it might be faster than CRV

    # obtain estimated fixed effects
    fit1.fixef()

    # demean data
    Y_hat = fit1.predict(newdata=data)
    _first_u = data[f"{yname}"].to_numpy().flatten() - Y_hat
    data[f"{yname}_hat"] = _first_u

    # intercept needs to be dropped by hand due to the presence of fixed effects in the first stage
    fit2 = cast(
        Feols,
        feols(
            _second_stage_full,
            data=data,
            vcov="iid",
            drop_intercept=True,
            weights=weights,
        ),
    )
    _second_u = fit2.resid()

    return fit2, _first_u, _second_u


def _did2s_vcov(
    data: pd.DataFrame,
    yname: str,
    first_stage: str,
    second_stage: str,
    treatment: str,
    first_u: np.ndarray,
    second_u: np.ndarray,
    cluster: str,
    weights: Optional[str] = None,
):
    """
    Variance-Covariance matrix for DID2S.

    Compute a variance covariance matrix for Gardner's 2-stage Difference-in-Differences
    Estimator.

    Parameters
    ----------
    data: pandas.DataFrame
        The DataFrame containing all variables.
    yname: str
        The name of the dependent variable.
    first_stage: str
        The formula for the first stage.
    second_stage: str
        The formula for the second stage.
    treatment: str
        The name of the treatment variable.
    first_u: np.ndarray
        The first stage residuals.
    second_u: np.ndarray
        The second stage residuals.
    cluster: str
        The name of the cluster variable.
    weights : Optional[str].
        Default is None. Weights for WLS estimation. If None, all observations
        are weighted equally. If a string, the name of the column in `data` that
        contains the weights.

    Returns
    -------
        A variance covariance matrix.
    """
    cluster_col = data[cluster]
    _, clustid = pd.factorize(cluster_col)

    _G = clustid.nunique()  # actually not used here, neither in did2s

    if weights is None:
        weights_array = np.repeat(1.0, data.shape[0])
    else:
        weights_array = np.sqrt(data[weights].to_numpy())

    # some formula parsing to get the correct formula for the first and second stage model matrix
    first_stage_x, first_stage_fe = first_stage.split("|")
    first_stage_fe_list = [f"C({i})" for i in first_stage_fe.split("+")]
    first_stage_fe_fml = "+".join(first_stage_fe_list)
    first_stage = f"{first_stage_x}+{first_stage_fe_fml}"

    second_stage = f"{second_stage}"

    # note for future Alex: intercept needs to be dropped! it is not as fixed
    # effects are converted to dummies, hence has_fixed checks are False

    FML1 = FixestFormulaParser(f"{yname} {first_stage}")
    FML2 = FixestFormulaParser(f"{yname} {second_stage}")
    FixestFormulaDict1 = FML1.FixestFormulaDict
    FixestFormulaDict2 = FML2.FixestFormulaDict

    mm_dict_first_stage = model_matrix_fixest(
        FixestFormula=next(iter(FixestFormulaDict1.values()))[0],
        data=data,
        weights=None,
        drop_singletons=False,
        drop_intercept=False,
    )
    X1 = cast(pd.DataFrame, mm_dict_first_stage.get("X"))

    mm_second_stage = model_matrix_fixest(
        FixestFormula=next(iter(FixestFormulaDict2.values()))[0],
        data=data,
        weights=None,
        drop_singletons=False,
        drop_intercept=True,
    )  # reference values not dropped, multicollinearity error
    X2 = cast(pd.DataFrame, mm_second_stage.get("X"))

    X1 = csr_matrix(X1.to_numpy() * weights_array[:, None])
    X2 = csr_matrix(X2.to_numpy() * weights_array[:, None])

    # Weight first and second stage residuals
    first_u *= weights_array
    second_u *= weights_array

    X10 = X1.copy().tocsr()  # type: ignore
    treated_rows = np.where(data[treatment], 0, 1)
    X10 = X10.multiply(treated_rows[:, None])

    X10X10 = X10.T.dot(X10)
    X2X1 = X2.T.dot(X1)
    X2X2 = X2.T.dot(X2)  # tocsc() to fix spsolve efficiency warning

    V = spsolve(X10X10.tocsc(), X2X1.T.tocsc()).T  # type: ignore

    k = X2.shape[1]
    vcov = np.zeros((k, k))

    X10 = X10.tocsr()
    X2 = X2.tocsr()  # type: ignore

    for (
        _,
        g,
    ) in enumerate(clustid):
        idx_g: np.ndarray = cluster_col.values == g
        X10g = X10[idx_g, :]
        X2g = X2[idx_g, :]
        first_u_g = first_u[idx_g]
        second_u_g = second_u[idx_g]

        W_g = X2g.T.dot(second_u_g) - V @ X10g.T.dot(first_u_g)
        score = spsolve(X2X2, W_g)
        if score.ndim == 1:
            score = score.reshape(-1, 1)
        cov_g = score.dot(score.T)

        vcov += cov_g

    return vcov, _G



import re
import warnings
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.stats import norm

from pyfixest.estimation.estimation import feols
from pyfixest.estimation.feols_ import Feols

from .did2s import DID


class SaturatedEventStudy(DID):
    """
    Saturated event study with cohort-specific effect curves.

    Attributes
    ----------
    data : pd.DataFrame
        Dataframe containing the data.
    yname : str
        Name of the outcome variable.
    idname : str
        Name of the unit identifier variable.
    tname : str
        Name of the time variable.
    gname : str
        Name of the treatment variable.
    cluster : str
        The name of the cluster variable.
    xfml : str
        Additional covariates to include in the model.
    att : bool
        Whether to use the average treatment effect.
    display_warning: bool
        Whether to display (some) warning messages.

    """

    def __init__(
        self,
        data: pd.DataFrame,
        yname: str,
        idname: str,
        tname: str,
        gname: str,
        att: bool = True,
        cluster: Optional[str] = None,
        xfml: Optional[str] = None,
        display_warning: bool = True,
    ):
        super().__init__(
            data=data,
            yname=yname,
            idname=idname,
            tname=tname,
            gname=gname,
            cluster=cluster,
            xfml=xfml,
            att=att,
        )
        self._estimator = "Saturated Event Study"

        if display_warning:
            warnings.warn(
                "The SaturatedEventStudyClass is currently in beta. Please report any issues you may encounter."
            )

    def estimate(self) -> Feols:
        """
        Estimate the model.

        Returns
        -------
        Feols
            The fitted Feols model object.
        """
        self.mod, self._res_cohort_eventtime_dict = _saturated_event_study(
            self._data,
            outcome=self._yname,
            time_id=self._tname,
            unit_id=self._idname,
            cluster=self._cluster,
        )

        return self.mod

    # !TODO - implement the rest of the methods
    def vcov(self):
        """
        Get the covariance matrix.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the covariance matrix.
        """
        pass

    def iplot(self):
        """Plot DID estimates."""
        cmp = plt.get_cmap("Set1")

        fig, ax = plt.subplots(figsize=(10, 6))

        for cohort, values in self._res_cohort_eventtime_dict.items():
            time = np.array(values["time"], dtype=float)
            est = values["est"]["Estimate"].astype(float).values
            ci_lower = values["est"]["2.5%"].astype(float).values
            ci_upper = values["est"]["97.5%"].astype(float).values

            ax.plot(time, est, marker="o", label=cohort, color=cmp(len(ax.lines)))
            ax.fill_between(
                time, ci_lower, ci_upper, alpha=0.3, color=cmp(len(ax.lines))
            )

        ax.axhline(0, color="black", linewidth=1, linestyle="--")
        ax.set_xlabel("Time")
        ax.set_ylabel("Coefficient (with 95% CI)")
        ax.set_title("Event Study Estimates by Cohort")
        ax.legend()
        plt.tight_layout()
        plt.show()

    def tidy(self):
        """Tidy result dataframe."""
        return self.mod.tidy()

    def summary(self):
        """Get summary table."""
        return self.mod.summary()

    def test_treatment_heterogeneity(self) -> pd.Series:
        """
        Test for treatment heterogeneity in the event study design.

        Parameters
        ----------
        by : str, optional

                The type of test to perform. Can be either "cohort" or "time".
                Default is "cohort". If "cohort", tests for treatment heterogeneity
                across cohorts as in Lal (2025). See https://arxiv.org/abs/2503.05125
                for details.
        """
        return _test_treatment_heterogeneity(
            model=self.mod if isinstance(self, SaturatedEventStudy) else self,
        )

    def aggregate(
        self, agg="period", weighting: Optional[str] = "shares"
    ) -> pd.DataFrame:
        """
        Aggregate the fully interacted event study estimates by relative time, cohort, and time.

        Parameters
        ----------
        agg : str, optional

                The type of aggregation to perform. Can be either "att" or "cohort" or "period".
                Default is "att". If "att", computes the average treatment effect on the treated.
                If "cohort", computes the average treatment effect by cohort. If "period",
                computes the average treatment effect by period.

        weighting : str, optional

                    The type of weighting to use. Can be either 'shares' or 'variance'.

        Returns
        -------
        pd.Series
            A Series containing the aggregated estimates.
        """
        if agg not in ["period"]:
            raise ValueError("agg must be either 'period'")

        if weighting not in ["shares"]:
            raise ValueError("weighting must be 'shares'.")

        model = self.mod if isinstance(self, SaturatedEventStudy) else self

        cohort_event_dict = model._res_cohort_eventtime_dict
        cohort_list = list(cohort_event_dict.keys())
        period_set = sorted(
            set(t for x in cohort_list for t in cohort_event_dict[x]["time"].tolist())
        )

        coefs = model._beta_hat
        se = model._se
        coefnames = model._coefnames

        if weighting == "shares":
            weights_df = compute_period_weights(
                data=model._data,
                cohort=model._gname,
                period="rel_time",
                treatment="is_treated",
            ).set_index([self._gname, "rel_time"])

        treated_periods = list(period_set)

        df_agg = pd.DataFrame(
            index=treated_periods,
            columns=["Estimate", "Std. Error", "t value", "Pr(>|t|)", "2.5%", "97.5%"],
        )
        df_agg.index.name = "period"

        for period in treated_periods:
            R = np.zeros(len(coefs))
            for cohort in cohort_list:
                cohort_pattern = rf"\[{re.escape(str(period))}\]:.*{re.escape(cohort)}$"
                match_idx = [
                    i
                    for i, name in enumerate(coefnames)
                    if re.search(cohort_pattern, name)
                ]
                cohort_int = int(cohort.replace("cohort_dummy_", ""))
                R[match_idx] = (
                    weights_df.xs((cohort_int, period)).values[0]
                    if weighting == "shares"
                    else 1 / se[match_idx]
                )

            if weighting == "variance":
                R /= np.sum(R)

            res_dict = _compute_lincomb_stats(R=R, coefs=coefs, vcov=model._vcov)
            df_agg.loc[period] = pd.Series(res_dict)

        return df_agg

    def iplot_aggregate(self, agg="period", weighting: Optional[str] = "shares"):
        """
        Plot the aggregated estimates.

        Parameters
        ----------
        agg : str, optional
            The type of aggregation to perform. Can be either "att" or "cohort" or "period".
            Default is "att". If "att", computes the average treatment effect on the treated.
            If "cohort", computes the average treatment effect by cohort. If "period",
            computes the average treatment effect by period.

        weighting : str, optional
            The type of weighting to use. Can be either 'shares' or 'variance'.

        Returns
        -------
        None
        """
        df_agg = self.aggregate(agg=agg, weighting=weighting)

        time = np.array(df_agg.index, dtype=float).astype(float)
        est = df_agg["Estimate"].values.astype(float)
        ci_lower = df_agg["2.5%"].values.astype(float)
        ci_upper = df_agg["97.5%"].values.astype(float)

        cmp = plt.get_cmap("Set1")
        fig, ax = plt.subplots(figsize=(10, 6))

        ax.plot(time, est, marker="o", color=cmp(len(ax.lines)))
        ax.fill_between(time, ci_lower, ci_upper, alpha=0.3, color=cmp(len(ax.lines)))

        ax.axhline(0, color="black", linewidth=1, linestyle="--")
        ax.set_xlabel("Time")
        ax.set_ylabel("Coefficient (with 95% CI)")
        ax.set_title("Event Study Estimates")
        ax.legend()
        plt.tight_layout()
        plt.show()


def _compute_lincomb_stats(R: np.ndarray, coefs: np.ndarray, vcov: np.ndarray) -> dict:
    """
    Compute linear combination of coefficients and statistics of interest.

    Parameters
    ----------
    R : np.ndarray
        1D numpy array of weights (same length as coefs),
    coefs: np.ndarray
        1D numpy array of coefficient estimates,
    vcov: np.ndarray
        The covariance matrix of coefs (same dimension as len(coefs)),

    Returns
    -------
    dict
        A dictionary containing the coefficient, standard error, t-statistic,
        p-value, and confidence interval bounds.
    """
    coef_val = np.sum(R * coefs)
    se = np.sqrt(R @ vcov @ R)
    tstat = coef_val / se
    pval = 2 * (1 - norm.cdf(abs(tstat)))
    z = norm.ppf(0.975)

    ci_margin = z * se
    conf_lower = coef_val - ci_margin
    conf_upper = coef_val + ci_margin

    return {
        "Estimate": coef_val,
        "Std. Error": se,
        "t value": tstat,
        "Pr(>|t|)": pval,
        "2.5%": conf_lower,
        "97.5%": conf_upper,
    }


def _saturated_event_study(
    df: pd.DataFrame,
    outcome: str,
    time_id: str,
    unit_id: str,
    cluster: Optional[str] = None,
):
    cohort_dummies = pd.get_dummies(
        df.first_treated_period, drop_first=True, prefix="cohort_dummy"
    )
    df_int = pd.concat([df, cohort_dummies], axis=1)

    ff = f"""
                {outcome} ~
                {"+".join([f"i(rel_time, {x}, ref = -1.0)" for x in cohort_dummies.columns.tolist()])}
                | {unit_id} + {time_id}
                """
    m = feols(fml=ff, data=df_int, vcov={"CRV1": cluster})  # type: ignore
    res = m.tidy()
    # create a dict with cohort specific effect curves
    res_cohort_eventtime_dict: dict[str, dict[str, pd.DataFrame | np.ndarray]] = {}
    for cohort in cohort_dummies.columns:
        res_cohort = res.filter(like=cohort, axis=0)
        event_time = (
            res_cohort.index.str.extract(r"\[(?:T\.)?(-?\d+(?:\.\d+)?)\]")
            .astype(float)
            .values.flatten()
        )
        res_cohort_eventtime_dict[cohort] = {"est": res_cohort, "time": event_time}

    return m, res_cohort_eventtime_dict


def _test_treatment_heterogeneity(
    model: Feols,
) -> pd.Series:
    """
    Test for treatment heterogeneity in the event study design.

    For details, see https://github.com/apoorvalal/TestingInEventStudies

    Parameters
    ----------
    model : SaturatedEventStudy
        The fitted event study model

    Returns
    -------
    pd.Series

            A Series containing the p-value of the test and the test statistic.
    """
    mmres = model.tidy().reset_index()
    P = mmres.shape[0]
    mmres[["time", "cohort"]] = mmres.Coefficient.str.split(":", expand=True)
    mmres["time"] = mmres.time.str.extract(r"\[(?:T\.)?(-?\d+(?:\.\d+)?)\]").astype(
        float
    )
    mmres["cohort"] = mmres.cohort.str.extract(r"(\d+)")
    # indices of coefficients that are deviations from common event study coefs
    event_study_coefs = mmres.loc[~(mmres.cohort.isna()) & (mmres.time > 0)].index
    # Method 2 (K x P) - more efficient
    K = len(event_study_coefs)
    R2 = np.zeros((K, P))
    for i, idx in enumerate(event_study_coefs):
        R2[i, idx] = 1

    test_result = model.wald_test(R=R2, distribution="chi2")
    return test_result


def compute_period_weights(
    data: pd.DataFrame,
    cohort: str = "g",
    period: str = "rel_time",
    treatment: str = "treatment",
    include_grid: bool = True,
) -> pd.DataFrame:
    """
    Compute Sun & Abraham interaction weights for all relative times.

    For l < 0, weight_{g,l} = n^0_{g,l} / Σ_{g'} n^0_{g',l}, where
      n^0_{g,l} = count of (cohort=g, period=l) still untreated but eventually treated.
    For l > 0, weight_{g,l} = n^1_{g,l} / Σ_{g'} n^1_{g',l}, where
      n^1_{g,l} = count of (cohort=g, period=l) already treated.
    For l = 0, weight is set to zero in the final grid.

    Parameters
    ----------
    data : pd.DataFrame
        Must include columns for `cohort`, `period`, and `treatment`.
    cohort : str
        Column name of cohort/adoption group.
    period : str
        Column name of relative time (lead/lag) indicator.
    treatment : str
        Column name of treatment indicator (0/1).
    include_grid : bool, default True
        If True, returns a full (cohort x period) grid with zero-filled weights.

    Returns
    -------
    pd.DataFrame
        Columns [cohort, period, weight]. If `include_grid`, every
        combination appears (with weight=0 where not defined).
    """
    df = data[[cohort, period, treatment]].copy()
    ever_treated = df.loc[df[treatment] == 1, cohort].unique()

    # post-treatment cells (l > 0)
    post = (
        df[df[treatment] == 1]
        .groupby([cohort, period])
        .size()
        .reset_index(name="n_grel")
    )
    post = post[post[period] >= 0]
    denom_post = post.groupby(period)["n_grel"].sum().reset_index(name="n_rel")
    post = post.merge(denom_post, on=period)
    post["weight"] = post["n_grel"] / post["n_rel"]

    # pre-treatment cells (l < 0)
    pre = (
        df[(df[treatment] == 0) & (df[cohort].isin(ever_treated))]
        .groupby([cohort, period])
        .size()
        .reset_index(name="n_grel")
    )
    pre = pre[pre[period] < 0]
    denom_pre = pre.groupby(period)["n_grel"].sum().reset_index(name="n_rel")
    pre = pre.merge(denom_pre, on=period)
    pre["weight"] = pre["n_grel"] / pre["n_rel"]

    # combine and (optionally) fill out the full grid
    out = pd.concat(
        [pre[[cohort, period, "weight"]], post[[cohort, period, "weight"]]],
        ignore_index=True,
    )

    if include_grid:
        all_periods = sorted(df[period].unique())
        grid = pd.MultiIndex.from_product(
            [list(ever_treated), all_periods], names=[cohort, period]
        ).to_frame(index=False)
        out = grid.merge(out, on=[cohort, period], how="left")
        out["weight"] = out["weight"].fillna(0.0)

    return out[[cohort, period, "weight"]]



import numpy as np
import pytest

from pyfixest.estimation.detect_singletons_ import detect_singletons
from pyfixest.estimation.jax.detect_singletons_jax import detect_singletons_jax

input1 = np.array([[0, 2, 1], [0, 2, 1], [0, 1, 3], [0, 1, 2], [0, 1, 2]])
solution1 = np.array([False, False, True, False, False])

input2 = np.array([[0, 2, 1], [0, 2, 1], [3, 1, 2], [0, 1, 1], [0, 1, 2]])
solution2 = np.array([False, False, True, True, True])

input3 = np.array([[0, 2, 1], [0, 2, 1], [0, 1, 1], [0, 1, 2], [0, 1, 2]])
solution3 = np.array([False, False, False, False, False])


@pytest.mark.parametrize(
    argnames="input, solution",
    argvalues=[(input1, solution1), (input2, solution2), (input3, solution3)],
)
@pytest.mark.parametrize(
    argnames="detection_function",
    argvalues=[detect_singletons, detect_singletons_jax],
    ids=["numba", "jax"],
)
def test_correctness(input, solution, detection_function):
    assert np.array_equal(detection_function(input), solution)


@pytest.mark.parametrize(
    argnames="detection_function",
    argvalues=[detect_singletons, detect_singletons_jax],
    ids=["numba", "jax"],
)
def test_single_column(detection_function):
    """Test with a single fixed effect column."""
    input_data = np.array([[0], [0], [1], [2], [2]])
    expected = np.array([False, False, True, False, False])
    result = detection_function(input_data)
    assert np.array_equal(result, expected)


@pytest.mark.parametrize(
    argnames="detection_function",
    argvalues=[detect_singletons, detect_singletons_jax],
    ids=["numba", "jax"],
)
def test_all_singletons(detection_function):
    """Test when all observations are singletons."""
    input_data = np.array([[0, 1], [1, 2], [2, 3], [3, 4]])
    expected = np.array([True, True, True, True])
    result = detection_function(input_data)
    assert np.array_equal(result, expected)


@pytest.mark.parametrize(
    argnames="detection_function",
    argvalues=[detect_singletons, detect_singletons_jax],
    ids=["numba", "jax"],
)
def test_no_singletons(detection_function):
    """Test when there are no singletons."""
    input_data = np.array([[0, 0], [0, 0], [1, 1], [1, 1]])
    expected = np.array([False, False, False, False])
    result = detection_function(input_data)
    assert np.array_equal(result, expected)


@pytest.mark.parametrize(
    argnames="detection_function",
    argvalues=[detect_singletons, detect_singletons_jax],
    ids=["numba", "jax"],
)
def test_large_input(detection_function):
    """Test with a larger input to check performance and correctness."""
    rng = np.random.default_rng(42)
    N = 10000
    input_data = np.column_stack(
        [
            rng.integers(0, N // 10, N),
            rng.integers(0, N // 5, N),
            rng.integers(0, N // 2, N),
        ]
    )

    # For large input, we compare against the Numba implementation as reference
    reference = detect_singletons(input_data)
    result = detection_function(input_data)

    assert np.array_equal(result, reference)
    assert len(result) == N
    assert result.dtype == np.bool_



import numba as nb
import numpy as np
from numba.extending import overload


def _prepare_fixed_effects(ary):
    pass


@overload(_prepare_fixed_effects)
def _ol_preproc_fixed_effects(ary):
    # If array is already an F-array we tolerate
    # any dtype because it saves us a copy
    if ary.layout == "F":
        return lambda ary: ary

    if not isinstance(ary.dtype, nb.types.Integer):
        raise nb.TypingError("Fixed effects must be integers")

    max_nbits = 32
    nbits = min(max_nbits, ary.dtype.bitwidth)
    dtype = nb.types.Integer.from_bitwidth(nbits, signed=False)

    def impl(ary):
        n, m = ary.shape
        out = np.empty((m, n), dtype=dtype).T
        out[:] = ary[:]
        return out

    return impl


@nb.njit
def detect_singletons(ids: np.ndarray) -> np.ndarray:
    """
    Detect singleton fixed effects in a dataset.

    This function iterates over the columns of a 2D numpy array representing
    fixed effects to identify singleton fixed effects.
    An observation is considered a singleton if it is the only one in its group
    (fixed effect identifier).

    Parameters
    ----------
    ids : np.ndarray
        A 2D numpy array representing fixed effects, with a shape of (n_samples,
        n_features).
        Elements should be non-negative integers representing fixed effect identifiers.

    Returns
    -------
    numpy.ndarray
        A boolean array of shape (n_samples,), indicating which observations have
        a singleton fixed effect.

    Notes
    -----
    The algorithm iterates over columns to identify fixed effects. After each
    column is processed, it updates the record of non-singleton rows. This approach
    accounts for the possibility that removing an observation in one column can
    lead to the emergence of new singletons in subsequent columns.

    For performance reasons, the input array should be in column-major order.
    Operating on a row-major array can lead to significant performance losses.
    """
    ids = _prepare_fixed_effects(ids)
    n_samples, n_features = ids.shape

    max_fixef = np.max(ids)
    counts = np.empty(max_fixef + 1, dtype=np.uint32)

    n_non_singletons = n_samples
    non_singletons = np.arange(n_non_singletons, dtype=np.uint32)

    while True:
        n_non_singletons_curr = n_non_singletons

        for j in range(n_features):
            col = ids[:, j]

            counts[:] = 0
            n_singletons = 0
            for i in range(n_non_singletons):
                e = col[non_singletons[i]]
                c = counts[e]
                # Branchless version of:
                #
                # if counts[e] == 1:
                #     n_singletons -= 1
                # elif counts[e] == 0:
                #     n_singletons += 1
                #
                n_singletons += (c == 0) - (c == 1)
                counts[e] += 1

            if not n_singletons:
                continue

            cnt = 0
            for i in range(n_non_singletons):
                e = col[non_singletons[i]]
                if counts[e] != 1:
                    non_singletons[cnt] = non_singletons[i]
                    cnt += 1

            n_non_singletons = cnt

        if n_non_singletons_curr == n_non_singletons:
            break

    is_singleton = np.ones(n_samples, dtype=np.bool_)
    for i in range(n_non_singletons):
        is_singleton[non_singletons[i]] = False

    return is_singleton



from typing import Optional, Union, cast

import numpy as np
import pandas as pd

from pyfixest.did.did import DID
from pyfixest.estimation.estimation import feols
from pyfixest.estimation.feols_ import Feols
from pyfixest.estimation.literals import VcovTypeOptions
from pyfixest.report.visualize import _HAS_LETS_PLOT, _coefplot


class LPDID(DID):
    """
    A class used to represent the Local Projections Diff-in-Diff Estimator.

    Attributes
    ----------
    data : pandas.DataFrame
        The DataFrame containing all variables.
    yname : str
        The name of the dependent variable.
    idname : str
        The name of the identifier variable.
    tname : str
        The name of the time variable.
    gname : str
        The name of the group variable.
    xfml : str
        The transformation to apply to the data.
    att : str
        The attribute to consider in the model.
    cluster : str
        The cluster to consider in the model.
    vcov : str
        The type of variance-covariance matrix to use.
    pre_window : tuple
        The pre-treatment window to consider in the model.
    post_window : tuple
        The post-treatment window to consider in the model.
    never_treated : bool
        Whether to consider never-treated units in the model.
    """

    def __init__(
        self,
        data: pd.DataFrame,
        yname: str,
        idname: str,
        tname: str,
        gname: str,
        xfml: str,
        att: bool,
        cluster: str,
        vcov: Optional[Union[VcovTypeOptions, dict[str, str]]] = None,
        pre_window: Optional[int] = None,
        post_window: Optional[int] = None,
        never_treated: Optional[int] = 0,
    ):
        # if att:
        #    raise NotImplementedError("ATT is not yet supported.")

        super().__init__(
            data=data,
            yname=yname,
            idname=idname,
            tname=tname,
            gname=gname,
            xfml=xfml,
            att=att,
            cluster=cluster,
        )
        assert isinstance(xfml, str) or xfml is None, "xfml must be a string or None"

        data = data.copy()
        data.sort_values([idname, tname], inplace=True)
        data[f"{yname}_lag"] = data.groupby(idname)[yname].shift(1)

        data["treat"] = np.where(data[gname] <= data[tname], 1, 0)
        data["rel_year"] = data[tname] - data[gname]

        # handle never treated units
        data["treat"] = np.where(data[gname] == never_treated, 0, data["treat"])
        data["rel_year"] = np.where(
            data[gname] == never_treated, np.inf, data["rel_year"]
        )

        data["treat_diff"] = data["treat"] - data.groupby(idname)["treat"].shift(1)
        data["treat_diff"] = np.where(data["treat_diff"] < 0, 0, data["treat_diff"])

        self._data = data

        rel_years = np.unique(data["rel_year"])
        rel_years = rel_years[np.isfinite(rel_years)]

        pre_window_int = int(np.min(rel_years)) if pre_window is None else pre_window
        post_window_int = int(np.max(rel_years)) if post_window is None else post_window

        # check that pre_window is in rel_years
        if pre_window_int not in rel_years:
            raise ValueError(f"pre_window must be in {rel_years}")
        # check that post_window is in rel_years
        if post_window_int not in rel_years:
            raise ValueError(f"post_window must be in {rel_years}")

        pre_window_int = np.abs(pre_window_int)

        if vcov is None:
            vcov = {"CRV1": idname}

        self._vcov = vcov
        self._pre_window = pre_window_int
        self._post_window = post_window_int
        self._never_treated = never_treated
        self._xfml = xfml
        self._estimator = "lpdid"

    def estimate(self):
        """Estimate the DID model."""
        self._coeftable = _lpdid_estimate(
            data=self._data,
            yname=self._yname,
            idname=self._idname,
            tname=self._tname,
            vcov=self._vcov,
            pre_window=self._pre_window,
            post_window=self._post_window,
            att=self._att,
            xfml=self._xfml,
        )

    def vcov(self):  # noqa: D102
        pass

    def iplot(
        self,
        alpha: float = 0.05,
        figsize: Optional[tuple[int, int]] = None,
        yintercept: Optional[int] = None,
        xintercept: Optional[int] = None,
        rotate_xticks: int = 0,
        title: str = "LPDID Event Study Estimate",
        coord_flip: bool = False,
        plot_backend: str = "lets_plot" if _HAS_LETS_PLOT else "matplotlib",
    ):
        """
        Create coefficient plots.

        Parameters
        ----------
        alpha : float, optional
            Significance level for visualization options. Defaults to 0.05.
        figsize : tuple[int, int], optional
            Size of the plot (width, height) in inches. Defaults to (500, 300).
        yintercept : float, optional
            Value to set as the y-axis intercept (vertical line).
            Defaults to None.
        xintercept : float, optional
            Value to set as the x-axis intercept (horizontal line).
            Defaults to None.
        rotate_xticks : int, optional
            Rotation angle for x-axis tick labels. Defaults to 0.
        title : str, optional
            Title of the plot.
        coord_flip : bool, optional
            Whether to flip the coordinates of the plot. Defaults to False.
        plot_backend: str, optional
            The plotting backend to use between "lets_plot" (default) and "matplotlib".

        Returns
        -------
        lets-plot figure
            A lets-plot or matplotlib figure with coefficient estimates and confidence intervals.
        """
        df = self._coeftable.copy().reset_index()
        df["fml"] = "lpdid"

        return _coefplot(
            plot_backend=plot_backend,
            df=df,
            figsize=figsize,
            alpha=alpha,
            yintercept=yintercept,
            xintercept=xintercept,
            rotate_xticks=rotate_xticks,
            title=title,
            flip_coord=coord_flip,
        )

    def tidy(self):  # noqa: D102
        return self._coeftable

    def summary(self):  # noqa: D102
        return self._coeftable


def _lpdid_estimate(
    data: pd.DataFrame,
    yname: str,
    idname: str,
    tname: str,
    pre_window: int,
    post_window: int,
    att: bool = True,
    vcov: Optional[Union[VcovTypeOptions, dict[str, str]]] = None,
    xfml: Optional[str] = None,
) -> pd.DataFrame:
    """
    Estimate a  Difference-in-Differences / Event Study Model via Linear Projections.

    Parameters
    ----------
    data: pandas.DataFrame
        The DataFrame containing all variables.
    yname: str
        The name of the dependent variable.
    idname: str
        The name of the id variable.
    tname: str
        Variable name for calendar period.
    gname:  str
        unit-specific time of initial treatment.
    vcov: VcovTypeOptions, dict[str, str], None
        The name of the cluster variable. If None, then defaults to {"CRV1": idname}.
        Either "iid", "hetero", or a dictionary, e.g. {"CRV1": idname} or
        {"CRV3": "idname"}. You can pass anything that is accepted by the vcov
        argument of feols.
    pre_window: int
        The number of periods before the treatment to include in the
        estimation. Default is None, which means that the pre_window is set to
        the minimum relative year in the data.
    post_window: int
        The number of periods after the treatment to include in the
        estimation. Default is None, which means that the post_window is set to
        the maximum relative year in the data.
    never_treated: int
        Value in gname that indicates that a unit was never treated.
        By default, never treated units are assumed to have value gname = 0.
    att: bool
        Whether to estimate the average treatment effect on the treated (ATT)
        or a canonical event study design with all leads and lags.
        Default is True.
    xfml: str
        Optional formula for the covariates. Not yet supported.
        E.g. "X1 + X2 + X3".

    Returns
    -------
    pandas.DataFrame
        A DataFrame with the estimated coefficients.
    """
    # the implementation here is highly influenced by Alex Cardazzi's R
    # code for the lpdid package: https://github.com/alexCardazzi/lpdid

    fit_all = []
    reweight = False  # noqa: F841

    if xfml is None:
        fml = f"Dy ~ treat_diff | {tname}"
    else:
        fml = f"Dy ~ treat_diff + {xfml} | {tname}"

    if att:
        # post window

        data[f"{yname}_post"] = _pooled_adjustment(data, yname, post_window, idname)
        data["Dy"] = data[f"{yname}_post"] - data[f"{yname}_lag"]
        sample_idx_post = (data["treat_diff"] == 1) | (
            data.groupby(idname)["treat"].shift(-post_window) == 0
        )
        fit_post = cast(Feols, feols(fml=fml, data=data[sample_idx_post], vcov=vcov))
        fit_tidy_post = fit_post.tidy().xs("treat_diff")
        fit_tidy_post["N"] = int(fit_post._N)

        res = pd.DataFrame(fit_tidy_post).T

    else:
        for h in range(post_window + 1):
            data["Dy"] = data.groupby(idname)[yname].shift(-h) - data[f"{yname}_lag"]

            sample_idx = (data["treat_diff"] == 1) | (
                data.groupby(idname)["treat"].shift(-h) == 0
            )

            fit = cast(Feols, feols(fml=fml, data=data[sample_idx], vcov=vcov))

            fit_tidy = fit.tidy().xs("treat_diff")
            fit_tidy["N"] = int(fit._N)
            fit_tidy.name = h
            fit_all.append(fit_tidy)

        for h in range(pre_window + 1):
            if h <= 1:
                # skip the reference period
                continue

            data["Dy"] = data.groupby(idname)[yname].shift(h) - data[f"{yname}_lag"]
            sample_idx = (data["treat_diff"] == 1) | (data["treat"] == 0)

            fit = cast(Feols, feols(fml=fml, data=data[sample_idx], vcov=vcov))

            fit_tidy = fit.tidy().xs("treat_diff")
            fit_tidy["N"] = int(fit._N)
            fit_tidy.name = -h
            fit_all.append(fit_tidy)

        res = pd.DataFrame(fit_all).sort_index()
        res.index.name = "Coefficient"
        res.index = res.index.map(lambda x: f"time_to_treatment::{x}")

    return res


def _pooled_adjustment(
    df: pd.DataFrame, y: str, pool_lead: int, idname: str
) -> np.ndarray:
    """
    Calculate post-treatment means rather than just using a single y value from t+h.

    Parameters
    ----------
    - df: pandas.DataFrame
        The dataset used in the analysis.
    - y: str
        The column name that denotes the outcome variable.
    - pool_lead: int
        The number of post-periods that should be used when calculating the
        post-treatment mean.
    - idname: str
        The name of the id variable.

    Returns
    -------
    np.ndarray
        The average of all future values in the analysis.
    """
    # Initialize lead variable
    x = np.zeros(df.shape[0])

    # Calculate lead sum
    for k in range(0, pool_lead + 1, 1):
        x += df.groupby(idname)[y].shift(-k).to_numpy()

    # Average the lead sum
    x /= pool_lead + 1

    return x



import re
from typing import Optional, Union

import narwhals.stable.v1 as nw
import numpy as np
import pandas as pd
from narwhals.typing import IntoDataFrame

DataFrameType = IntoDataFrame


def _narwhals_to_pandas(data: IntoDataFrame) -> pd.DataFrame:  # type: ignore
    return nw.from_native(data, eager_or_interchange_only=True).to_pandas()


def _create_rng(seed: Optional[int] = None) -> np.random.Generator:
    """
    Create a random number generator.

    Parameters
    ----------
    seed : int, optional
        The seed of the random number generator. If None, a random seed is chosen.

    Returns
    -------
    numpy.random.Generator
        A random number generator.
    """
    if seed is None:
        seed = np.random.randint(100_000_000)
    return np.random.default_rng(seed)


def _select_order_coefs(
    coefs: list,
    keep: Optional[Union[list, str]] = None,
    drop: Optional[Union[list, str]] = None,
    exact_match: Optional[bool] = False,
):
    r"""
    Select and order the coefficients based on the pattern.

    Parameters
    ----------
    coefs: list
        Coefficient names to be selected and ordered.
    keep: str or list of str, optional
        The pattern for retaining coefficient names. You can pass a string (one
        pattern) or a list (multiple patterns). Default is keeping all coefficients.
        You should use regular expressions to select coefficients.
            "age",            # would keep all coefficients containing age
            r"^tr",           # would keep all coefficients starting with tr
            r"\\d$",          # would keep all coefficients ending with number
        Output will be in the order of the patterns.
    drop: str or list of str, optional
        The pattern for excluding coefficient names. You can pass a string (one
        pattern) or a list (multiple patterns). Syntax is the same as for `keep`.
        Default is keeping all coefficients. Parameter `keep` and `drop` can be
        used simultaneously.
    exact_match: bool, optional
        Whether to use exact match for `keep` and `drop`. Default is False.
        If True, the pattern will be matched exactly to the coefficient name
        instead of using regular expressions.

    Returns
    -------
    res: list
        The filtered and ordered coefficient names.
    """
    if keep is None:
        keep = []
    if drop is None:
        drop = []

    if isinstance(keep, str):
        keep = [keep]
    if isinstance(drop, str):
        drop = [drop]

    coefs = list(coefs)
    res = [] if keep else coefs[:]  # Store matched coefs
    for pattern in keep:
        _coefs = []  # Store remaining coefs
        for coef in coefs:
            if (exact_match and pattern == coef) or (
                exact_match is False and re.findall(pattern, coef)
            ):
                res.append(coef)
            else:
                _coefs.append(coef)
        coefs = _coefs

    for pattern in drop:
        _coefs = []
        for coef in res:  # Remove previously matched coefs that match the drop pattern
            if (exact_match and pattern == coef) or (
                exact_match is False and re.findall(pattern, coef)
            ):
                continue
            else:
                _coefs.append(coef)
        res = _coefs

    return res


def docstring_from(func, custom_doc=""):
    """Copy the docstring of another function."""

    def decorator(target_func):
        target_func.__doc__ = custom_doc + "\n\n" + func.__doc__
        return target_func

    return decorator


def _check_series_or_dataframe(x: Union[pd.Series, pd.DataFrame]):
    if not isinstance(x, (pd.Series, pd.DataFrame)):
        raise TypeError("Input must be a pandas Series or DataFrame")
    else:
        return x


def _to_list(x):
    if x is not None and not isinstance(x, list):
        return [x]
    return x


def _drop_cols(_data: pd.DataFrame, na_index: np.ndarray):
    """
    Drop columns from data based on the indices in na_index.

    Parameters
    ----------
    _data : pd.DataFrame
        The input DataFrame.
    na_index : np.ndarray
        An array of indices to drop.

    Returns
    -------
    pd.DataFrame
        The input DataFrame with NAs dropped.
    """
    if na_index.size > 0:
        all_indices = np.arange(_data.shape[0])
        max_index = all_indices.max() + 1
        keep = np.ones(max_index, dtype=bool)
        keep[na_index] = False
        return _data[keep]
    else:
        return _data


def _extract_variable_level(fe_string: str):
    """
    Extract the variable and level from a given string.

    Parameters
    ----------
    fe_string: str
        The string encapsulating the fixed effect factor variable and level.

    Returns
    -------
    tuple
        A tuple containing the extracted variable and level for the fixed
        effect.
    """
    pattern = r"C\(([^)]*)\)\[(?:T\.)?(.*)\]$"
    match = re.search(pattern, fe_string)
    if not match:
        raise ValueError(f"Cannot parse: {fe_string}")

    variable = match.group(1)
    level = match.group(2)

    return f"C({variable})", level
